from django.contrib import admin

from .models import App
from .models import Authorization
from .models import Event
from .models import Node
from .models import Process
from .models import ProcessCategory
from .models import ProcessInstance
from .models import ProcessReportLink
from .models import Task
from .models import Transition


class ProcessCategoryAdmin(admin.ModelAdmin):
    search_fields = ('name', )
    list_display = ('name', 'oid', 'is_active')


class ProcessReportLinkAdmin(admin.ModelAdmin):
    search_fields = ('category__name', 'name', 'url')
    list_display = ('name', 'url', 'category', 'perm', 'oid', 'is_active')
    list_filter = ('category',)


class ProcessAdmin(admin.ModelAdmin):
    search_fields = ('code', 'prefix', 'name', 'category__name')
    list_display = ('code', 'prefix', 'name', 'category', 'oid', 'is_active')
    list_filter = ('category',)


class NodeAdmin(admin.ModelAdmin):
    search_fields = (
        'process__name', 'process__code', 'name', 'code',
        'operators', 'notice_users', 'share_users')
    list_display = (
        'process', 'name', 'code', 'step', 'status', 'audit_page_type',
        'can_edit', 'can_reject', 'can_give_up',
        'operators', 'notice_users', 'share_users',
        'is_active')
    list_filter = ('process',)


class TransitionAdmin(admin.ModelAdmin):
    search_fields = (
        'process__name', 'process__code',
        'input_node__name', 'output_node__name',
        'name', 'condition')
    list_display = (
        'process', 'name', 'code', 'routing_rule',
        'input_node', 'output_node',
        'is_agree', 'can_auto_agree',
        'app', 'app_param', 'condition',
        'oid', 'is_active')
    list_filter = ('process',)


class AppAdmin(admin.ModelAdmin):
    list_display = ('name', 'app_type', 'action')


class ProcessInstanceAdmin(admin.ModelAdmin):
    search_fields = (
        'process__name', 'process__code',
        'created_by__username', 'cur_node__name')
    list_display = (
        'process', 'no', 'summary', 'created_by', 'created_on',
        'cur_node')
    list_filter = ('process', )
    raw_id_fields = (
        'content_type', 'created_by', 'attachments', 'can_view_users',
        'cur_node')


class TaskAdmin(admin.ModelAdmin):
    search_fields = (
        'instance__no', 'node__name', 'user__username',
        'agent__username')
    list_display = (
        'instance', 'node', 'user', 'agent_user', 'is_hold', 'status',
        'created_on', 'receive_on')
    list_filter = ('instance__process', )
    raw_id_fields = ('instance', 'node', 'user', 'agent_user', 'authorization', )


class EventAdmin(admin.ModelAdmin):
    search_fields = (
        'instance__no', 'user__username', 'old_node__name',
        'new_node__name', )
    list_display = (
        'instance', 'user', 'get_act_name', 'old_node',
        'new_node', 'created_on')
    raw_id_fields = (
        'instance', 'user', 'task', 'next_operators',
        'notice_users')


class AuthorizationAdmin(admin.ModelAdmin):
    search_fields = ('user__username', 'agent_user__username')
    list_display = ('user', 'agent_user', 'start_on', 'end_on')
    raw_id_fields = ('user', 'agent_user')


admin.site.register(ProcessCategory, ProcessCategoryAdmin)
admin.site.register(Process, ProcessAdmin)
admin.site.register(Node, NodeAdmin)
admin.site.register(Transition, TransitionAdmin)
admin.site.register(App, AppAdmin)
admin.site.register(ProcessInstance, ProcessInstanceAdmin)
admin.site.register(Task, TaskAdmin)
admin.site.register(Event, EventAdmin)
admin.site.register(Authorization, AuthorizationAdmin)
admin.site.register(ProcessReportLink, ProcessReportLinkAdmin)



# Generated by Django 2.2.10 on 2020-02-21 06:06

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('lbworkflow', '0003_auto_20200221_0438'),
    ]

    operations = [
        migrations.CreateModel(
            name='SimpleWorkFlow',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_on', models.DateTimeField(auto_now_add=True, verbose_name='Created on')),
                ('summary', models.CharField(max_length=255, verbose_name='Summary')),
                ('content', models.TextField(blank=True, verbose_name='Content')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='Created by')),
                ('pinstance', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='simpleworkflow', to='lbworkflow.ProcessInstance', verbose_name='Process instance')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from lbworkflow.core.datahelper import create_category
from lbworkflow.core.datahelper import create_node
from lbworkflow.core.datahelper import create_process
from lbworkflow.core.datahelper import create_transition


def load_data():
    load_simplewf()


def load_simplewf():
    category = create_category('5f31d065-00cc-0020-beea-641f0a670010', 'HR')

    process = create_process('simplewf__A', 'Simple Workflow: A', category=category)
    create_node('5f31d666-00a0-0020-beea-641f0a670010', process, 'Draft', status='draft')
    create_node('5f31d666-00a0-0020-beea-641f0a670020', process, 'Given up', status='given up')
    create_node('5f31d666-00a0-0020-beea-641f0a670030', process, 'Rejected', status='rejected')
    create_node('5f31d666-00a0-0020-beea-641f0a670040', process, 'Completed', status='completed')
    create_node('5f31d666-00a0-0020-beea-641f0a670050', process, 'A1', operators='[owner]')
    create_transition('5f31d666-00e0-0020-beea-641f0a670010', process, 'Draft,', 'A1')
    create_transition('5f31d666-00e0-0020-beea-641f0a670020', process, 'A1,', 'Completed')

    process = create_process('simplewf__B', 'Simple Workflow: B', category=category)
    create_node('5f31d667-00a0-0020-beea-641f0a670010', process, 'Draft', status='draft')
    create_node('5f31d667-00a0-0020-beea-641f0a670020', process, 'Given up', status='given up')
    create_node('5f31d667-00a0-0020-beea-641f0a670030', process, 'Rejected', status='rejected')
    create_node('5f31d667-00a0-0020-beea-641f0a670040', process, 'Completed', status='completed')
    create_node('5f31d667-00a0-0020-beea-641f0a670050', process, 'A1', operators='[owner]')
    create_transition('5f31d667-00e0-0020-beea-641f0a670010', process, 'Draft,', 'A1')
    create_transition('5f31d667-00e0-0020-beea-641f0a670020', process, 'A1,', 'Completed')



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.db import models
from lbutils import get_or_none

from lbworkflow.core.helper import safe_eval

User = get_user_model()


def remove_brackets(s, start_char="[", end_char="]"):
    return s.strip(start_char).strip(end_char).strip()


class BaseUserParser(object):

    def __init__(self, param, pinstance=None, operator=None, owner=None):
        self.owner = owner
        self.operator = operator
        self.param = param
        self.pinstance = pinstance
        self.wf_obj = None
        if pinstance:
            self.wf_obj = pinstance.content_object
            self.owner = pinstance.created_by

    def _get_eval_val(self, eval_str):
        return safe_eval(eval_str, {'o': self.wf_obj})

    def eval_as_list(self, eval_str):
        v = self._get_eval_val(eval_str)
        if v.__class__.__name__ == 'ManyRelatedManager':
            return v.all()
        if isinstance(v, models.Model):
            return [v]
        return v

    def parse(self):
        return []


class SimpleUserParser(BaseUserParser):
    def process_func(self, func_str):
        return None

    def _get_users(self, user_str):
        """
        #owner
        #operator
        [11:vicalloy]
        [o.auditor]
        [o.auditors]
        """
        user_str = remove_brackets(user_str)
        if user_str.startswith('#'):
            user_str = user_str[1:]
            if user_str == 'owner':
                return [self.owner]
            elif user_str == 'operator':
                return [self.operator]
        if '.' in user_str:
            return self.eval_as_list(user_str)
        if ':' in user_str:
            pk = user_str.split(':')[0]
            return User.objects.filter(pk=pk)
        return User.objects.filter(username=user_str)

    def _get_groups(self, group_str):
        """
        g[o.group]
        g[o.groups]
        g[11:admins]
        """
        group_str = remove_brackets(group_str, 'g[')
        if '.' in group_str:
            return self.eval_as_list(group_str)
        pk = group_str.split(':')[0]
        group = get_or_none(Group, pk=pk)
        return [group] if group else []

    def _get_users_by_groups(self, group_str):
        groups = self._get_groups(group_str)
        return User.objects.filter(group__in=groups)

    def _paser_atom_rule(self, atom_rule):
        """
        #owner
        #operator
        user [11:vicalloy]
        group g[11:group]

        if syntax error will return None
        """
        if not atom_rule:
            return []
        users = self.process_func(atom_rule)
        if users is not None:  # is function
            return users
        if atom_rule.startswith('#'):
            return self._get_users(atom_rule)
        elif atom_rule.startswith('g['):  # role(group)
            return self._get_users_by_groups(atom_rule)
        elif atom_rule.startswith('['):  # user
            return self._get_users(atom_rule)
        # log it?
        return None

    def _to_users(self, rules):
        all_users = []
        for rule in rules:
            users = self._paser_atom_rule(rule)
            if users is not None:
                all_users.extend(users)
        # TODO ignore quited users
        return all_users

    def _get_active_rules(self):
        """
        :o.leave_days<7
        [vicalloy]
        :o.leave_days>=7
        [tom]
        """
        rules = [e.strip() for e in self.param.splitlines() if e.strip()]
        str_rules = ''
        need_add = True
        for rule in rules:
            is_condition = rule.startswith(':')
            if not is_condition and not need_add:
                continue
            if is_condition:
                need_add = safe_eval(rule[1:], {'o': self.wf_obj})
                continue
            str_rules = '%s,%s' % (str_rules, rule)
        return [e.strip() for e in str_rules.split(',') if e.strip()]

    def parse(self):
        rules = self._get_active_rules()
        active_rules = []
        for rule in rules:
            active_rules.append(rule)
        users = self._to_users(active_rules)
        return list(set(users))



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-07-07 08:00
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('lbattachment', '0001_initial'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='App',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('name', models.CharField(help_text='Name', max_length=100)),
                ('app_type', models.CharField(choices=[('url', 'URL')], default='url', max_length=255, verbose_name='Type')),
                ('action', models.CharField(blank=True, help_text="URL: It can be url or django's url name. If it's blank will use transition's app param", max_length=255)),
                ('note', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Authorization',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_on', models.DateField(verbose_name='Start on')),
                ('end_on', models.DateField(verbose_name='End on')),
                ('agent_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='agent_user_authorizations', to=settings.AUTH_USER_MODEL, verbose_name='Agent user')),
            ],
        ),
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('act_type', models.CharField(choices=[('transition', 'Transition'), ('edit', 'Edit'), ('give up', 'Give up'), ('reject', 'Reject'), ('back to', 'Back to'), ('rollback', 'Rollback'), ('comment', 'Comment'), ('assign', 'Assign'), ('hold', 'Hold'), ('unhold', 'Unhold')], default='transition', max_length=255)),
                ('comment', models.TextField(blank=True, default='')),
                ('ext_data', jsonfield.fields.JSONField(blank=True, null=True)),
                ('created_on', models.DateTimeField(auto_now=True)),
                ('attachments', models.ManyToManyField(blank=True, to='lbattachment.LBAttachment', verbose_name='Attachment')),
            ],
        ),
        migrations.CreateModel(
            name='Node',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('code', models.CharField(blank=True, max_length=255, verbose_name='Code')),
                ('step', models.IntegerField(default=0, verbose_name='Step')),
                ('status', models.CharField(choices=[('draft', 'Draft'), ('given up', 'Given up'), ('rejected', 'Rejected'), ('in progress', 'In Progress'), ('completed', 'Completed')], default='in progress', max_length=16, verbose_name='Type')),
                ('audit_page_type', models.CharField(choices=[('view', 'view'), ('edit', 'Edit')], default='view', help_text='If this node can edit, will auto goto edit mode when audit.', max_length=64, verbose_name='Audit page type')),
                ('can_edit', models.BooleanField(default=False, verbose_name='Can edit')),
                ('can_reject', models.BooleanField(default=True, verbose_name='Can reject')),
                ('can_give_up', models.BooleanField(default=True, verbose_name='Can give up')),
                ('operators', models.TextField(blank=True, verbose_name='Audit users')),
                ('notice_users', models.TextField(blank=True, verbose_name='Notice users')),
                ('share_users', models.TextField(blank=True, verbose_name='Share users')),
                ('note', models.TextField(blank=True, verbose_name='Note')),
                ('is_active', models.BooleanField(default=True, verbose_name='Is active')),
                ('ext_data', jsonfield.fields.JSONField(default='{}')),
            ],
        ),
        migrations.CreateModel(
            name='Process',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(help_text='A unique code to identify process', max_length=100, unique=True, verbose_name='Code')),
                ('prefix', models.CharField(blank=True, default='', help_text='prefix for process NO.', max_length=8, verbose_name='Prefix')),
                ('name', models.CharField(help_text='Name for this process', max_length=255, verbose_name='Name')),
                ('note', models.TextField(blank=True, verbose_name='Note')),
                ('oid', models.IntegerField(default=999, verbose_name='Order')),
                ('is_active', models.BooleanField(default=True, verbose_name='Is active')),
                ('ext_data', jsonfield.fields.JSONField(default='{}')),
            ],
            options={
                'verbose_name': 'Process',
                'ordering': ['oid'],
                'permissions': (('sft_mgr_process', 'workflow - Config'),),
            },
        ),
        migrations.CreateModel(
            name='ProcessCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('name', models.CharField(db_index=True, max_length=255, verbose_name='Name')),
                ('oid', models.IntegerField(default=999, verbose_name='Order')),
                ('is_active', models.BooleanField(default=True, verbose_name='Is active')),
            ],
            options={
                'ordering': ['oid'],
            },
        ),
        migrations.CreateModel(
            name='ProcessInstance',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('no', models.CharField(blank=True, max_length=100, verbose_name='NO.')),
                ('object_id', models.PositiveIntegerField()),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('submit_on', models.DateTimeField(blank=True, null=True)),
                ('end_on', models.DateTimeField(blank=True, null=True)),
                ('summary', models.TextField(blank=True, verbose_name='Summary')),
                ('attachments', models.ManyToManyField(blank=True, to='lbattachment.LBAttachment')),
                ('can_view_users', models.ManyToManyField(blank=True, related_name='can_view_pinstances', to=settings.AUTH_USER_MODEL, verbose_name='Can view users')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='instances', to=settings.AUTH_USER_MODEL)),
                ('cur_node', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='lbworkflow.Node')),
                ('process', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='lbworkflow.Process')),
            ],
        ),
        migrations.CreateModel(
            name='ProcessReportLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('url', models.CharField(max_length=255, verbose_name='URL')),
                ('open_in_new_window', models.BooleanField(default=False, verbose_name='Open in new window')),
                ('perm', models.CharField(blank=True, help_text='Permission to view this report', max_length=255, verbose_name='Permission')),
                ('oid', models.IntegerField(default=999, verbose_name='Order')),
                ('is_active', models.BooleanField(default=True, verbose_name='Is active')),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='lbworkflow.ProcessCategory', verbose_name='Category')),
            ],
            options={
                'ordering': ['oid'],
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(choices=[('in progress', 'In Progress'), ('completed', 'Completed')], default='in progress', max_length=255)),
                ('receive_on', models.DateTimeField(blank=True, null=True, verbose_name='Receive on')),
                ('is_hold', models.BooleanField(default=False, verbose_name='Is hold')),
                ('created_on', models.DateTimeField(auto_now=True)),
                ('agent_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='agent_user_tasks', to=settings.AUTH_USER_MODEL, verbose_name='Agent user')),
                ('authorization', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='lbworkflow.Authorization', verbose_name='Authorization')),
                ('instance', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='lbworkflow.ProcessInstance')),
                ('node', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='lbworkflow.Node')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
        ),
        migrations.CreateModel(
            name='Transition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('name', models.CharField(default='Agree', help_text="It also the action's name, like: Agree/Submit", max_length=100, verbose_name='Name')),
                ('code', models.CharField(blank=True, max_length=100, verbose_name='Code')),
                ('is_agree', models.BooleanField(default=True, help_text='User only need agree one time in one workflow', verbose_name='Is agree')),
                ('can_auto_agree', models.BooleanField(default=True, help_text='If user agreed in previous steps will auto agree', verbose_name='If can auto agree')),
                ('routing_rule', models.CharField(choices=[('split', 'split'), ('joint', 'Joint')], default='split', help_text='joint: do transition after all work item finished. joint: do transition immediately', max_length=16, verbose_name='Routing rule')),
                ('app_param', models.CharField(blank=True, help_text='Depend on App config', max_length=100, null=True, verbose_name='Param for application')),
                ('condition', models.TextField(blank=True, help_text='Uses the Python syntax.ex: `o.leave_days>3`', verbose_name='Condition')),
                ('note', models.TextField(blank=True, verbose_name='Note')),
                ('oid', models.IntegerField(default=999, verbose_name='Order')),
                ('is_active', models.BooleanField(default=True, verbose_name='Is active')),
                ('ext_data', jsonfield.fields.JSONField(default='{}')),
                ('app', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='lbworkflow.App', verbose_name='Application to perform')),
                ('input_node', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='input_transitions', to='lbworkflow.Node', verbose_name='Input node')),
                ('output_node', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='output_transitions', to='lbworkflow.Node', verbose_name='Output node')),
                ('process', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='lbworkflow.Process', verbose_name='Process')),
            ],
        ),
        migrations.AddField(
            model_name='process',
            name='category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='lbworkflow.ProcessCategory', verbose_name='Category'),
        ),
        migrations.AddField(
            model_name='node',
            name='process',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='lbworkflow.Process', verbose_name='Process'),
        ),
        migrations.AddField(
            model_name='event',
            name='instance',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='lbworkflow.ProcessInstance'),
        ),
        migrations.AddField(
            model_name='event',
            name='new_node',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='in_events', to='lbworkflow.Node'),
        ),
        migrations.AddField(
            model_name='event',
            name='next_operators',
            field=models.ManyToManyField(blank=True, related_name='audit_events', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='event',
            name='notice_users',
            field=models.ManyToManyField(blank=True, related_name='notice_events', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='event',
            name='old_node',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='out_events', to='lbworkflow.Node'),
        ),
        migrations.AddField(
            model_name='event',
            name='task',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='events', to='lbworkflow.Task'),
        ),
        migrations.AddField(
            model_name='event',
            name='transition',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='lbworkflow.Transition'),
        ),
        migrations.AddField(
            model_name='event',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='authorization',
            name='processes',
            field=models.ManyToManyField(to='lbworkflow.Process', verbose_name='Processes'),
        ),
        migrations.AddField(
            model_name='authorization',
            name='user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='authorized_user_authorizations', to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-21 04:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lbworkflow', '0002_auto_20171019_0549'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='is_joint',
            field=models.BooleanField(default=False, verbose_name='Is joint'),
        ),
        migrations.AlterField(
            model_name='event',
            name='created_on',
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='created_on',
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



from django.core.exceptions import ImproperlyConfigured
from django.forms import ModelForm
from django.http import HttpResponseRedirect
from django.views.generic.base import ContextMixin
from django.views.generic.base import View

try:
    from crispy_forms.helper import FormHelper
except ImportError:
    pass


__all__ = ('FormsMixin', 'ModelFormsMixin', 'FormSetMixin', 'FormsView', 'BSFormSetMixin')


class FormsMixin(ContextMixin):
    """
    A mixin that provides a way to show and handle any number of form in a request.
    """

    initial = {}
    form_classes = None  # the main form should named as form
    success_url = None

    def get_initial(self, form_class_key):
        return self.initial.get(form_class_key, {}).copy()

    def get_form_classes(self):
        """
        Returns the form classes to use in this view
        """
        if not self.form_classes:
            raise ImproperlyConfigured("Provide form_classes.")
        return self.form_classes

    def after_create_form(self, form_class_key, form):
        return form

    def create_form(self, form_class_key, form_class):
        form = form_class(**self.get_form_kwargs(form_class_key, form_class))
        self.after_create_form(form_class_key, form)
        return form

    def create_forms(self, **form_classes):
        """
        Returns an instance of the forms to be used in this view.
        forms can be access by self.forms
        """
        forms = {}
        self.forms = forms
        for form_class_key, form_class in form_classes.items():
            forms[form_class_key] = self.create_form(form_class_key, form_class)
        return forms

    def get_form_kwargs(self, form_class_key, form_class):
        """
        Returns the keyword arguments for instantiating the form.
        """
        kwargs = {'initial': self.get_initial(form_class_key)}
        if form_class_key != 'form':
            kwargs['prefix'] = form_class_key
        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        return kwargs

    def get_success_url(self):
        """
        Returns the supplied success URL.
        """
        if self.success_url:
            # Forcing possible reverse_lazy evaluation
            url = self.success_url
        else:
            raise ImproperlyConfigured(
                "No URL to redirect to. Provide a success_url.")
        return url

    def forms_valid(self, **forms):
        """
        If the forms are valid, redirect to the supplied URL.
        """
        return HttpResponseRedirect(self.get_success_url())

    def forms_invalid(self, **forms):
        """
        If the forms are invalid, re-render the context data with the
        data-filled form and errors.
        """
        return self.render_to_response(self.get_context_data(**forms))


class ModelFormsMixin:

    def get_form_kwargs(self, form_class_key, form_class):
        kwargs = super().get_form_kwargs(form_class_key, form_class)
        # not (ModelForm or ModelFormSet)
        formset_form_class = getattr(form_class, 'form', str)
        if not issubclass(form_class, ModelForm) \
                and not issubclass(formset_form_class, ModelForm):
            return kwargs
        instance = getattr(self, 'object', None)
        # if have main form, try to get instance from main form
        # other form may have ForeignKey to main object
        form = self.forms.get('form')
        if form and getattr(form, 'instance', None):
            instance = getattr(form, 'instance', None)
        kwargs['instance'] = instance
        return kwargs


def is_formset(form):
    # form class
    if getattr(form, '__name__', '').endswith('FormSet'):
        return True
    # form instance
    return type(form).__name__.endswith('FormSet')


class FormSetMixin:

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        formset_list = []
        for form in self.forms.values():
            if is_formset(form):
                formset_list.append(form)
        kwargs['formset_list'] = formset_list
        return kwargs

    def after_create_formset(self, form_class_key, formset):
        formset.title = "Items"

    def after_create_form(self, form_class_key, form):
        super().after_create_form(form_class_key, form)
        if is_formset(form):
            self.after_create_formset(form_class_key, form)
        return form

    def get_formset_kwargs(self, form_class_key, form_class):
        return {}

    def get_form_kwargs(self, form_class_key, form_class):
        kwargs = super().get_form_kwargs(form_class_key, form_class)
        if is_formset(form_class):
            return kwargs
        ext_kwargs = self.get_formset_kwargs(form_class_key, form_class)
        kwargs.update(ext_kwargs)
        return kwargs


class FormsView(FormSetMixin, ModelFormsMixin, FormsMixin, View):
    """
    A mixin that renders any number of forms on GET and processes it on POST.
    """
    def get(self, request, *args, **kwargs):
        """
        Handles GET requests and instantiates a blank version of the forms.
        """
        form_classes = self.get_form_classes()
        forms = self.create_forms(**form_classes)
        return self.render_to_response(self.get_context_data(**forms))

    def post(self, request, *args, **kwargs):
        """
        Handles POST requests, instantiating a form instance with the passed
        POST variables and then checked for validity.
        """
        form_classes = self.get_form_classes()
        forms = self.create_forms(**form_classes)
        if all([forms[form].is_valid() for form in forms]):
            return self.forms_valid(**forms)
        else:
            return self.forms_invalid(**forms)

    # PUT is a valid HTTP verb for creating (with a known URL) or editing an
    # object, note that browsers only support POST for now.
    def put(self, *args, **kwargs):
        return self.post(*args, **kwargs)


class BSFormSetMixin:
    """
    Crispy & Bootstrap for formset
    """
    def after_create_formset(self, form_class_key, formset):
        super().after_create_formset(form_class_key, formset)
        helper = FormHelper()
        helper.template = 'lbadminlte/bootstrap3/table_inline_formset.html'
        formset.helper = helper
        return formset



from django.urls import path

from .views import flowchart
from .views import processinstance
from .views.list import ListWF
from .views.list import MyWF
from .views.list import Todo
from .views.transition import AddAssigneeView
from .views.transition import BatchExecuteAgreeTransitionView
from .views.transition import BatchExecuteGiveUpTransitionView
from .views.transition import BatchExecuteRejectTransitionView
from .views.transition import ExecuteAgreeTransitionView
from .views.transition import ExecuteBackToTransitionView
from .views.transition import ExecuteGiveUpTransitionView
from .views.transition import ExecuteRejectTransitionView
from .views.transition import ExecuteTransitionView
from .views.transition import execute_transitions

urlpatterns = [
    path('t/', ExecuteTransitionView.as_view(), name="wf_execute_transition"),
    path('t/agree/', ExecuteAgreeTransitionView.as_view(), name="wf_agree"),
    path('t/back_to/', ExecuteBackToTransitionView.as_view(), name="wf_back_to"),
    path('t/reject/', ExecuteRejectTransitionView.as_view(), name="wf_reject"),
    path('t/give_up/', ExecuteGiveUpTransitionView.as_view(), name="wf_give_up"),
    path('t/add_assignee/', AddAssigneeView.as_view(), name="wf_add_assignee"),
    path('t/batch/agree/', BatchExecuteAgreeTransitionView.as_view(), name="wf_batch_agree"),
    path('t/batch/reject/', BatchExecuteRejectTransitionView.as_view(), name="wf_batch_reject"),
    path('t/batch/give_up/', BatchExecuteGiveUpTransitionView.as_view(), name="wf_batch_give_up"),
    path(
        't/e/<str:wf_code>/<str:trans_func>/', execute_transitions,
        name='wf_execute_transition'),

    path('start_wf/', processinstance.start_wf, name='wf_start_wf'),
    path('report_list/', processinstance.report_list, name='wf_report_list'),

    path('new/<str:wf_code>/', processinstance.new, name='wf_new'),
    path('delete/', processinstance.delete, name='wf_delete'),
    path('list/<str:wf_code>/', processinstance.show_list, name='wf_list'),
    path('edit/<int:pk>/', processinstance.edit, name='wf_edit'),
    path('<int:pk>/', processinstance.detail, name='wf_detail'),
    path('<int:pk>/print/', processinstance.detail,
         {
             'ext_ctx': {'is_print': True}
         },
         name='wf_print_detail'),
    path('todo/', Todo.as_view(), name='wf_todo'),
    path('list/', ListWF.as_view(), name='wf_list_wf'),
    path('my/', MyWF.as_view(), name='wf_my_wf'),

    path('flowchart/process/<str:wf_code>/',
         flowchart.process_flowchart,
         name='wf_process_flowchart'),
    path('flowchart/<int:pk>/',
         flowchart.process_instance_flowchart,
         name='wf_process_instance_flowchart'),
]



from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.generic.base import TemplateResponseMixin
from django.views.generic.base import View
from django.views.generic.list import MultipleObjectMixin
from django.views.generic.list import MultipleObjectTemplateResponseMixin
from lbutils import do_filter
from lbutils import simple_export2xlsx

from lbworkflow import settings
from lbworkflow.forms import BSQuickSearchForm
from lbworkflow.forms import BSQuickSearchWithExportForm
from lbworkflow.models import Process

from .forms import FormsView
from .helper import get_base_wf_permit_query_param
from .helper import get_wf_template_names
from .helper import user_wf_info_as_dict


class WorkflowTemplateResponseMixin(TemplateResponseMixin):
    def get_template_names(self):
        try:
            return super().get_template_names()
        except ImproperlyConfigured:
            return get_wf_template_names(
                self.wf_code,
                self.base_template_name,
                wf_object=getattr(self, 'object', None),
                model=getattr(self, 'model', None))


class ExcelResponseMixin:
    excel_titles = []
    excel_file_name = 'flow'

    def get_excel_data(self, o):
        return []

    def get_formated_excel_data(self, o):
        data = []
        for d in self.get_excel_data(o):
            # TODO date
            d = "%s" % d
            data.append(d)
        return data

    def render_to_excel(self, object_list, **kwargs):
        return simple_export2xlsx(
            self.excel_file_name, self.excel_titles,
            object_list, lambda o: self.get_formated_excel_data(o))


class CreateView(WorkflowTemplateResponseMixin, FormsView):
    form_classes = {
        # 'form': None,  # the form for BaseWFObj should named as form
    }
    wf_code = None
    model = None
    base_template_name = 'form.html'

    def get_success_url(self):
        return reverse('wf_detail', args=(self.object.pinstance.pk, ))

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['wf_code'] = self.wf_code
        kwargs['process'] = get_object_or_404(Process, code=self.wf_code)
        return kwargs

    def forms_valid(self, **forms):
        form = forms.pop('form')
        self.object = form.save_new_process(self.request, self.wf_code)
        # TODO refactor, you may update total amount base items
        for form in forms.values():
            form.save()
        return HttpResponseRedirect(self.get_success_url())

    def dispatch(self, request, wf_code, *args, **kwargs):
        self.wf_code = wf_code
        return super().dispatch(request, *args, **kwargs)


class UpdateView(WorkflowTemplateResponseMixin, FormsView):
    form_classes = {
        # 'form': None,  # the form for BaseWFObj should named as form
    }
    wf_code = None
    model = None
    base_template_name = 'form.html'

    def get_success_url(self):
        return reverse('wf_detail', args=(self.object.pinstance.pk, ))

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs.update(user_wf_info_as_dict(self.object, self.request.user))
        return kwargs

    def forms_valid(self, **forms):
        form = forms.pop('form')
        self.object = form.update_process(self.request)
        # TODO refactor
        for form in forms.values():
            form.save()
        return HttpResponseRedirect(self.get_success_url())

    def dispatch(self, request, wf_object, *args, **kwargs):
        self.object = wf_object
        self.wf_code = wf_object.pinstance.process.code
        return super().dispatch(request, *args, **kwargs)


class BaseListView(ExcelResponseMixin, MultipleObjectMixin, View):
    paginate_by = settings.WF_PAGE_SIZE
    search_form_class = BSQuickSearchForm
    quick_query_fields = []
    int_quick_query_fields = []
    ordering = '-pk'
    base_template_name = 'list.html'

    def dispatch(self, request, *args, wf_code=None, **kwargs):
        self.wf_code = wf_code
        return super().dispatch(request, *args, **kwargs)

    def get_quick_query_fields(self):
        return self.quick_query_fields

    def do_filter(self, queryset, query_data):
        quick_query_fields = self.get_quick_query_fields()
        return do_filter(queryset, query_data, quick_query_fields, self.int_quick_query_fields)

    def get_search_form(self, request):
        if not self.search_form_class:
            return None
        search_form = self.search_form_class(request.GET)
        if not search_form.is_valid():
            pass
        return search_form

    def get(self, request, *args, **kwargs):
        search_form = self.get_search_form(request)
        queryset = self.get_queryset()
        self.object_list = self.do_filter(
            queryset,
            search_form.cleaned_data if search_form else {})

        if 'export' in request.GET:
            return self.render_to_excel(self.object_list)

        process = None
        if self.wf_code:
            process = get_object_or_404(Process, code=self.wf_code)
        context = self.get_context_data(
            search_form=search_form, process=process)
        return self.render_to_response(context)


class ListView(MultipleObjectTemplateResponseMixin, BaseListView):
    pass


class WFListView(WorkflowTemplateResponseMixin, BaseListView):
    search_form_class = BSQuickSearchWithExportForm
    model = None
    wf_code = None

    def get_quick_query_fields(self):
        fields = [
            'pinstance__no',
            'pinstance__summary',
            'pinstance__created_by__username',
            'pinstance__cur_node__name',
        ]
        fields.extend(self.quick_query_fields)
        return fields

    def get_permit_query_param(self, user, q_param):
        # override this function to add addition permit
        return q_param

    def get_queryset(self):
        qs = super().get_queryset()
        # only show have permission
        user = self.request.user
        if self.wf_code:
            qs = qs.filter(pinstance__process__code=self.wf_code)
        if not user.is_superuser:
            return qs
        q_param = get_base_wf_permit_query_param(user)
        q_param = self.get_permit_query_param(user, q_param)
        qs = qs.filter(q_param).distinct()
        return qs



from django.db import models

from lbworkflow.models import BaseWFObj


class SimpleWorkFlow(BaseWFObj):
    summary = models.CharField('Summary', max_length=255)
    content = models.TextField('Content', blank=True)

    def __str__(self):
        return self.summary



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.utils import timezone
from lbutils import get_or_none

from lbworkflow.core.datahelper import load_wf_data

from .leave.models import Leave
from .wfdata import init_users

User = get_user_model()


class BaseTests(TestCase):

    def setUp(self):
        self.init_data()

    def init_users(self):
        super().setUp()
        self.users = init_users()

    # TODO add a function to submit new leave
    def create_leave(self, reason, submit=True):
        leave = Leave(
            start_on=timezone.now(), end_on=timezone.now(), leave_days=1,
            reason=reason, created_by=self.users['owner'])
        leave.init_actual_info()
        leave.save()
        leave.create_pinstance('leave', submit)
        return leave

    def get_leave(self, reason):
        return get_or_none(Leave, reason=reason)

    def init_leave(self):
        self.leave = self.create_leave('reason', False)

    def init_data(self):
        self.init_users()
        load_wf_data('lbworkflow')
        load_wf_data('lbworkflow.tests.leave')
        self.init_leave()



from django.utils import timezone

from lbworkflow.models import Event
from lbworkflow.models import Task

from .sendmsg import wf_send_msg


def create_event(instance, transition, **kwargs):
    act_type = 'transition' if transition.pk else transition.code
    if transition.is_agree:
        act_type = 'agree'
    event = Event.objects.create(
        instance=instance, act_name=transition.name, act_type=act_type,
        **kwargs)
    return event


class TransitionExecutor(object):
    def __init__(
            self, operator, instance, task, transition=None,
            comment='', attachments=[]):
        self.wf_obj = instance.content_object
        self.instance = instance
        self.operator = operator
        self.task = task
        self.transition = transition

        self.comment = comment
        self.attachments = attachments

        self.from_node = instance.cur_node
        # hold&assign wouldn't change node
        self.to_node = transition.output_node
        self.all_todo_tasks = instance.get_todo_tasks()

        self.last_event = None

    def execute(self):
        # TODO check permission

        all_todo_tasks = self.all_todo_tasks
        need_transfer = False
        if self.transition.code in ['reject', 'back to', 'give up']:
            need_transfer = True
        elif self.transition.routing_rule == 'joint':
            if all_todo_tasks.count() == 1:
                need_transfer = True
        else:
            if not all_todo_tasks.exclude(pk=self.task.pk).filter(is_joint=True).exists():
                need_transfer = True
        self._complete_task(need_transfer)
        if not need_transfer:
            return

        self._do_transfer()

        # if is agree should check if need auto agree for next node
        if self.transition.is_agree or self.to_node.node_type == 'router':
            self._auto_agree_next_node()

    def _auto_agree_next_node(self):
        instance = self.instance

        agree_transition = instance.get_agree_transition()
        all_todo_tasks = instance.get_todo_tasks()

        if not agree_transition:
            return

        # if from router, create a task
        if self.to_node.node_type == 'router':
            task = Task(
                instance=self.instance,
                node=self.instance.cur_node,
                user=self.operator,
            )
            all_todo_tasks = [task]

        for task in all_todo_tasks:
            users = [task.user, task.agent_user]
            users = [e for e in users if e]
            for user in set(users):
                if self.instance.cur_node != task.node:  # has processed
                    return
                if instance.is_user_agreed(user):
                    TransitionExecutor(self.operator, instance, task, agree_transition).execute()

    def _complete_task(self, need_transfer):
        """ close workite, create event and return it """
        instance = self.instance
        task = self.task
        transition = self.transition

        task.status = 'completed'
        task.save()

        to_node = self.to_node if need_transfer else instance.cur_node
        self.to_node = to_node

        event = None
        pre_last_event = instance.last_event()
        if pre_last_event and pre_last_event.new_node.node_type == 'router':
            event = pre_last_event
            event.new_node = to_node
            event.save()

        if not event:
            event = create_event(
                instance, transition,
                comment=self.comment, user=self.operator,
                old_node=task.node, new_node=to_node,
                task=task)

        if self.attachments:
            event.attachments.add(*self.attachments)

        self.last_event = event

        return event

    def _do_transfer_for_instance(self):
        instance = self.instance
        wf_obj = self.wf_obj

        from_node = self.from_node
        from_status = from_node.status

        to_node = self.to_node
        to_status = self.to_node.status

        # Submit
        if not from_node.is_submitted() and to_node.is_submitted():
            instance.submit_time = timezone.now()
            wf_obj.on_submit()

        # cancel & give up & reject
        if from_node.is_submitted() and not to_node.is_submitted():
            wf_obj.on_fail()

        # complete
        if from_status != 'completed' and to_status == 'completed':
            instance.end_on = timezone.now()
            self.wf_obj.on_complete()

        # cancel complete
        if from_status == 'completed' and to_status != 'completed':
            instance.end_on = None

        instance.cur_node = self.to_node
        self.wf_obj.on_do_transition(from_node, to_node)

        instance.save()

    def _send_notification(self):
        instance = self.instance
        last_event = self.last_event

        notice_users = last_event.notice_users.exclude(
            pk__in=[self.operator.pk, instance.created_by.pk]).distinct()
        wf_send_msg(notice_users, 'notify', last_event)

        # send notification to instance.created_by
        if instance.created_by != self.operator:
            wf_send_msg([instance.created_by], 'transfered', last_event)

    def _gen_new_task(self):
        last_event = self.last_event

        if not last_event:
            return

        next_operators = last_event.next_operators.distinct()

        need_notify_operators = []
        for operator in next_operators:
            new_task = Task(
                instance=self.instance, node=self.to_node,
                user=operator)
            new_task.update_authorization(commit=True)

            # notify next operator(not include current operator and instance.created_by)
            if operator not in [self.operator, self.instance.created_by]:
                need_notify_operators.append(operator)

            agent_user = new_task.agent_user
            if agent_user and agent_user not in [self.operator, self.instance.created_by]:
                need_notify_operators.append(agent_user)

        wf_send_msg(need_notify_operators, 'new_task', last_event)

    def update_users_on_transfer(self):
        instance = self.instance
        event = self.last_event
        to_node = event.new_node

        next_operators = to_node.get_operators(instance.created_by, self.operator, instance)
        event.next_operators.add(*next_operators)
        notice_users = to_node.get_notice_users(instance.created_by, self.operator, instance)
        event.notice_users.add(*notice_users)
        can_view_users = to_node.get_share_users(instance.created_by, self.operator, instance)
        instance.can_view_users.add(*can_view_users)

    def _do_transfer(self):
        self.update_users_on_transfer()
        # auto complete all current work item
        self.all_todo_tasks.update(status='completed')
        self._do_transfer_for_instance()
        self._gen_new_task()
        self._send_notification()



from django.db.models import Q
from django.utils import timezone
from lbutils import as_callable

from lbworkflow.models import ProcessInstance
from lbworkflow.models import Task
from lbworkflow.settings import PROCESS_INSTANCE_GET_PERMIT_QUERY_PARAM_FUNC
from lbworkflow.views.generics import ListView

from .helper import get_base_wf_permit_query_param


class ListWF(ListView):
    model = ProcessInstance
    ordering = '-created_on'
    template_name = 'lbworkflow/list_wf.html'
    search_form_class = None  # can config search_form_class
    quick_query_fields = [
        'no',
        'summary',
        'created_by__username',
        'cur_node__name',
    ]

    def get_permit_query_param(self, user, q_param):
        # override this function to add addition permit
        get_permit_query_param = as_callable(PROCESS_INSTANCE_GET_PERMIT_QUERY_PARAM_FUNC)
        return get_permit_query_param(user, q_param)

    def get_queryset(self):
        user = self.request.user
        qs = super().get_queryset()
        qs = qs.exclude(cur_node__status__in=['draft', 'given up'])
        if not user.is_superuser:
            q_param = get_base_wf_permit_query_param(user, '')
            q_param = self.get_permit_query_param(user, q_param)
            qs = qs.filter(q_param)
        qs = qs.select_related(
            'process',
            'created_by',
            'cur_node'
        ).distinct()
        return qs


class MyWF(ListView):
    model = ProcessInstance
    template_name = 'lbworkflow/my_wf.html'
    search_form_class = None  # can config search_form_class
    quick_query_fields = [
        'no',
        'summary',
        'cur_node__name',
    ]

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(created_by=self.request.user)


class Todo(ListView):
    model = Task
    template_name = 'lbworkflow/todo.html'
    search_form_class = None  # can config search_form_class
    quick_query_fields = [
        'instance__no',
        'instance__summary',
        'instance__cur_node__name',
        'instance__created_by__username',
    ]

    def get_queryset(self):
        user = self.request.user
        qs = super().get_queryset()
        qs = qs.filter(
            Q(user=user) | Q(agent_user=user),
            status='in progress')
        qs.filter(receive_on=None).update(receive_on=timezone.now())
        qs = qs.select_related(
            'instance',
            'instance__process',
            'instance__cur_node'
        ).distinct()
        return qs



from django.contrib.auth import get_user_model

from lbworkflow.core.userparser import SimpleUserParser

from .test_base import BaseTests

User = get_user_model()


class UserSimpleParserTests(BaseTests):

    def test_parser_users(self):
        users = SimpleUserParser('[vicalloy]').parse()
        self.assertEqual(users[0], self.users['vicalloy'])

        users = SimpleUserParser(
            '[%s:vicalloy]' % self.users['vicalloy'].pk
        ).parse()
        self.assertEqual(users[0], self.users['vicalloy'])

        users = SimpleUserParser('#owner', owner=self.users['owner']).parse()
        self.assertEqual(users[0], self.users['owner'])

        users = SimpleUserParser(
            '#operator', operator=self.users['operator']
        ).parse()
        self.assertEqual(users[0], self.users['operator'])

    def test_eval_as_list(self):
        # [o.auditors]
        users = SimpleUserParser('[o.created_by]', self.leave.pinstance).parse()
        self.assertEqual(users[0], self.users['owner'])

    def test_condition_rules(self):
        rules = """
        :True
        [owner]
        [operator]
        :False
        [vicalloy]
        """
        users = SimpleUserParser(
            rules,
            operator=self.users['operator'],
            owner=self.users['owner'],
        ).parse()
        self.assertEqual(
            set(users), set([self.users['owner'], self.users['operator']])
        )

    def test_parser_groups(self):
        # TODO
        pass



from django import template

register = template.Library()


@register.filter
def app_url(transition, task):
    return transition.get_app_url(task)


@register.filter
def flow_status_css_class(pinstance):
    if not pinstance:
        return 'default'
    if pinstance.cur_node.status in ['rejected']:
        return 'danger'
    if pinstance.cur_node.status == 'in progress':
        return 'info'
    if pinstance.cur_node.status == 'finished':
        return 'success'
    return 'default'


@register.filter
def category_have_perm_processes(category, user):
    return category.get_can_apply_processes(user)


@register.filter(is_safe=True)
def mermaid_transition_line(transition, event_transitions):
    if (transition.input_node, transition.output_node) in event_transitions:
        return '-->'
    return '-.->'



from lbworkflow.core.datahelper import create_category
from lbworkflow.core.datahelper import create_node
from lbworkflow.core.datahelper import create_process
from lbworkflow.core.datahelper import create_transition


def load_data():
    load_leave()


def load_leave():
    """ load_[wf_code] """
    category = create_category('5f31d065-00cc-0010-beea-641f0a670010', 'HR')
    process = create_process('leave', 'Leave', category=category)
    create_node('5f31d065-00a0-0010-beea-641f0a670010', process, 'Draft', status='draft')
    create_node('5f31d065-00a0-0010-beea-641f0a670010', process, 'Draft', status='draft')  # test for update
    create_node('5f31d065-00a0-0010-beea-641f0a670020', process, 'Given up', status='given up')
    create_node('5f31d065-00a0-0010-beea-641f0a670030', process, 'Rejected', status='rejected')
    create_node('5f31d065-00a0-0010-beea-641f0a670040', process, 'Completed', status='completed')
    create_node('5f31d065-00a0-0010-beea-641f0a670050', process, 'A1', operators='[owner]')
    create_node('5f31d065-00a0-0010-beea-641f0a670060', process, 'A2', operators='[tom]')
    create_node('5f31d065-00a0-0010-beea-641f0a670065', process, 'A2B1', operators='[tom],[owner]')
    create_node('5f31d065-00a0-0010-beea-641f0a670070', process, 'A3', operators='[vicalloy]')
    create_node('5f31d065-00a0-0010-beea-641f0a670080', process, 'A4', operators='[hr]')
    create_node('5f31d065-00a0-0010-beea-641f0a670a10', process, 'R1', operators='', node_type='router')
    create_transition('5f31d065-00e0-0010-beea-641f0a670010', process, 'Draft,', 'A1')
    create_transition('5f31d065-00e0-0010-beea-641f0a670020', process, 'A1,', 'A2')
    create_transition(
        '5f31d065-00e0-0010-beea-641f0a670030', process, 'A2,', 'A3',
        condition='o.leave_days<7  # days<7')
    create_transition(
        '5f31d065-00e0-0010-beea-641f0a670040', process, 'A2,', 'A2B1',
        condition='o.leave_days>=7  # days>=7')
    create_transition(
        '5f31d065-00e0-0010-beea-641f0a670050', process, 'A2B1,', 'A3',
        routing_rule='joint', can_auto_agree=False)
    create_transition('5f31d065-00e0-0010-beea-641f0a670060', process, 'A3,', 'A4')
    create_transition(
        '5f31d065-00e0-0010-beea-641f0a670070', process, 'A4,', 'R1',
        app='Customized URL', app_param='wf_execute_transition {{wf_code}} c')
    create_transition('5f31d065-00e0-0010-beea-641f0a670080', process, 'R1,', 'Completed')



from django.db import models

from lbworkflow.models import BaseWFObj


class Leave(BaseWFObj):
    start_on = models.DateTimeField('Start on')
    end_on = models.DateTimeField('End on')
    leave_days = models.DecimalField('Leave days', max_digits=5, decimal_places=1)

    actual_start_on = models.DateTimeField('Actual start on')
    actual_end_on = models.DateTimeField('Actual end on')
    actual_leave_days = models.DecimalField(
        'Actual leave days', max_digits=5, decimal_places=1)

    reason = models.TextField('Reason')

    class Meta:
        verbose_name = 'Leave'
        ordering = ["-created_on"]
        permissions = (
        )

    def __str__(self):
        return '%s %s days' % (self.created_by, self.leave_days, )

    def init_actual_info(self):
        self.actual_start_on = self.start_on
        self.actual_end_on = self.end_on
        self.actual_leave_days = self.leave_days



from django.contrib import admin
from django.urls import include
from django.urls import path
from django.views.generic import RedirectView

urlpatterns = [
    path('', RedirectView.as_view(url='/wf/list/'), name='home'),
    path('admin/', admin.site.urls),
    path('wf/', include('lbworkflow.urls')),
    path('attachment/', include('lbattachment.urls')),
    path('select2/', include('django_select2.urls')),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2017-10-19 05:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lbworkflow', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='event',
            name='transition',
        ),
        migrations.AddField(
            model_name='event',
            name='act_name',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='node',
            name='node_type',
            field=models.CharField(choices=[('node', 'Node'), ('router', 'Router')], default='node', max_length=16, verbose_name='Status'),
        ),
        migrations.AlterField(
            model_name='event',
            name='act_type',
            field=models.CharField(choices=[('transition', 'Transition'), ('agree', 'Agree'), ('edit', 'Edit'), ('give up', 'Give up'), ('reject', 'Reject'), ('back to', 'Back to'), ('rollback', 'Rollback'), ('comment', 'Comment'), ('assign', 'Assign'), ('hold', 'Hold'), ('unhold', 'Unhold')], default='transition', max_length=255),
        ),
        migrations.AlterField(
            model_name='node',
            name='status',
            field=models.CharField(choices=[('draft', 'Draft'), ('given up', 'Given up'), ('rejected', 'Rejected'), ('in progress', 'In Progress'), ('completed', 'Completed')], default='in progress', max_length=16, verbose_name='Status'),
        ),
    ]



from django.template import Context
from django.template import Template

from lbworkflow.models import Process
from lbworkflow.models import ProcessInstance


def get_event_transitions(process_instance):
    from lbworkflow.models import Event
    events = Event.objects.filter(instance=process_instance).order_by('-created_on', '-id')
    transitions = []
    for event in events:
        transition = (event.old_node, event.new_node)
        if event.new_node.status not in ['in progress']:
            transitions.append(transition)
            break
        if transition not in transitions:
            transitions.append(transition)
    return transitions


def generate_mermaid_src(process_instance):
    file_template = """
    {% load lbworkflow_tags %}
    graph TD
    {% for node in nodes %}
    {% if node.node_type == 'router' %}
    {{node.pk}}{ {{node.name}} }
    {% else %}
    {{node.pk}}( {{node.name}} )
    {% endif %}
    {% endfor %}
    {% for transition in transitions %}
    {{ transition.input_node.pk }} {{ transition|mermaid_transition_line:event_transitions|safe }}{% if transition.get_condition_descn %}|{{transition.get_condition_descn}}|{% endif %} {{ transition.output_node.pk }}
    {% endfor %}
    """  # NOQA
    if isinstance(process_instance, Process):
        process = process_instance
        process_instance = None
    else:
        process = process_instance.process

    transitions = process.transition_set.all()
    event_transitions = []
    if process_instance:
        event_transitions = get_event_transitions(process_instance)

    nodes = process.node_set.all()
    ctx = Context(
        {
            'name': process.name,
            'nodes': nodes,
            'transitions': transitions,
            'event_transitions': event_transitions
        }
    )
    t = Template(file_template)
    return t.render(ctx)


def process_flowchart(request, wf_code):
    from django.shortcuts import render
    template_name = 'lbworkflow/flowchart.html'
    process = Process.objects.get(code=wf_code)
    graph_src = generate_mermaid_src(process)
    ctx = {
        'process': process,
        'graph_src': graph_src
    }
    return render(request, template_name, ctx)


def process_instance_flowchart(request, pk):
    from django.shortcuts import render
    template_name = 'lbworkflow/flowchart.html'
    process_instance = ProcessInstance.objects.get(pk=pk)
    graph_src = generate_mermaid_src(process_instance)
    ctx = {
        'process': process_instance.process,
        'graph_src': graph_src
    }
    return render(request, template_name, ctx)



from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect
from django.shortcuts import render
from django.urls import reverse
from lbutils import as_callable

from lbworkflow import settings
from lbworkflow.models import Process
from lbworkflow.models import ProcessCategory
from lbworkflow.models import ProcessInstance

from .helper import get_wf_template_names
from .helper import import_wf_views
from .helper import user_wf_info_as_dict

can_edit_wf = as_callable(settings.CAN_EDIT_WF_FUNC)
can_submit_wf = as_callable(settings.CAN_SUBMIT_WF_FUNC)
can_view_wf = as_callable(settings.CAN_VIEW_WF_FUNC)


def new(request, wf_code):
    views = import_wf_views(wf_code)
    process = Process.objects.get(code=wf_code)
    if not can_submit_wf(process, request.user):
        raise PermissionDenied
    return views.new(request, wf_code=wf_code)


def show_list(request, wf_code):
    views = import_wf_views(wf_code)
    return views.show_list(request, wf_code=wf_code)


def edit(request, pk):
    instance = get_object_or_404(ProcessInstance, pk=pk)
    wf_code = instance.process.code
    views = import_wf_views(wf_code)
    if not can_edit_wf(instance, request.user):
        messages.info(request, 'No permission: %s' % instance)
        return redirect(reverse('wf_my_wf'))
    return views.edit(request, instance.content_object)


def _default_detail(request, instance, ext_ctx={}, template_name=None):
    is_print = ext_ctx.get('is_print')
    if not template_name:
        base_template_name = 'detail.html'
        if is_print:
            base_template_name = 'print.html'
        template_name = get_wf_template_names(
            instance.process.code,
            base_template_name,
            wf_object=instance.content_object)
    ctx = {}
    ctx.update(ext_ctx)
    return render(request, template_name, ctx)


def detail(request, pk, template_name=None, ext_ctx={}):
    instance = ProcessInstance.objects.get(pk=pk)
    views = import_wf_views(instance.process.code)
    is_print = ext_ctx.get('is_print')
    __ext_param_process = getattr(views, '__ext_param_process', None)
    ctx = {}
    ctx.update(ext_ctx)

    if not can_view_wf(instance, request.user, ext_param_process=__ext_param_process):
        messages.info(request, 'No permission to view this process')
        return redirect(reverse('wf_my_wf'))

    user_wf_info = user_wf_info_as_dict(instance.content_object, request.user)
    ctx.update(user_wf_info)

    if not is_print and instance.cur_node.can_edit \
            and instance.cur_node.audit_view_type == 'edit' \
            and ext_ctx['task'] and instance.cur_node.resolution == 'started':
        return redirect(reverse('wf_edit', args=[instance.pk]))
    detail_func = getattr(views, 'detail', _default_detail)
    ret = detail_func(request, instance, ext_ctx=ctx)
    if isinstance(ret, dict):
        ret.update(ctx)
        return _default_detail(request, instance, ret, template_name)
    return ret


def delete(request):
    pks = request.POST.getlist('pk') or request.GET.getlist('pk')
    instances = ProcessInstance.objects.filter(pk__in=pks)
    for instance in instances:
        # only workflow admin can delete
        if instance.is_wf_admin(request.user):
            instance.delete()
    messages.info(request, 'Deleted')
    return redirect(reverse('wf_list_wf'))


def start_wf(request):
    template_name = 'lbworkflow/start_wf.html'
    categories = ProcessCategory.objects.filter(
        is_active=True).order_by('oid')
    # only have perm's categories
    categories = [e for e in categories if e.get_can_apply_processes(request.user)]
    ctx = {
        'categories': categories,
    }
    return render(request, template_name, ctx)


def report_list(request):
    template_name = 'lbworkflow/report_list.html'
    categories = ProcessCategory.objects.filter(
        is_active=True).order_by('oid')
    categories = [e for e in categories]
    ctx = {
        'categories': categories,
    }
    return render(request, template_name, ctx)



import datetime

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q
from jsonfield import JSONField
from lbattachment.models import LBAttachment
from lbutils import get_or_none

from lbworkflow.settings import AUTH_USER_MODEL
from lbworkflow.settings import GET_USER_DISPLAY_NAME_FUNC

from .config import Node
from .config import Process
from .config import Transition


class ProcessInstance(models.Model):
    """
    A process instance is created when someone decides to do something,
    and doing this thing means start using a process defined in ``django-lb-workflow``.
    """
    no = models.CharField('NO.', max_length=100, blank=True)
    process = models.ForeignKey(
        'lbworkflow.Process',
        on_delete=models.CASCADE
    )
    created_by = models.ForeignKey(
        AUTH_USER_MODEL,
        null=True, on_delete=models.SET_NULL,
        related_name='instances')

    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')

    created_on = models.DateTimeField(auto_now_add=True)
    submit_on = models.DateTimeField(null=True, blank=True)
    end_on = models.DateTimeField(null=True, blank=True)

    cur_node = models.ForeignKey(
        Node,
        null=True, on_delete=models.SET_NULL
    )

    attachments = models.ManyToManyField(LBAttachment, blank=True)
    can_view_users = models.ManyToManyField(
        AUTH_USER_MODEL, blank=True,
        verbose_name='Can view users',
        related_name="can_view_pinstances")

    summary = models.TextField('Summary', blank=True)

    def __str__(self):
        return self.no

    def last_event(self):
        return self.event_set.order_by('-created_on', '-pk').first()

    def can_rollback(self, user):
        """ if can roll back, return last event """
        last_event = self.last_event()
        if not last_event:
            return None
        if last_event.new_node == self.cur_node \
                and self.cur_node.status == 'in progress' \
                and last_event.old_node.status == 'in progress' \
                and last_event.user == user \
                and last_event.act_type in ['reject', 'back to', 'transition', 'give up']:
            return last_event
        return None

    def is_wf_admin(self, user):
        # TODO permission
        return user.is_superuser

    def can_view(self, user, ext_param_process=None):
        # TODO permission
        return True

    def can_give_up(self, user):
        if self.cur_node.status != 'in progress':
            return False
        if not self.cur_node.can_give_up:
            return False
        if self.is_wf_admin(user):
            return True
        if self.created_by == user:
            return True
        return False

    def get_operators(self):
        # TODO select_related
        qs = self.task_set.filter(status='in progress')
        users = []
        for e in qs:
            users.extend([e.user, e.agent_user])
        return [e for e in users if e]

    def get_operators_display(self):
        return ', '.join([GET_USER_DISPLAY_NAME_FUNC(e) for e in self.get_operators()])

    def get_reject_transition(self):
        return self.process.get_reject_transition(self.cur_node)

    def get_back_to_transition(self, out_node=None):
        return self.process.get_back_to_transition(self.cur_node, out_node)

    def get_rollback_transition(self, out_node):
        return self.process.get_rollback_transition(self.cur_node, out_node)

    def get_give_up_transition(self):
        return self.process.get_give_up_transition(self.cur_node)

    def get_add_assignee_transition(self):
        return self.process.get_add_assignee_transition(self.cur_node)

    def create_task(self, operator, **kwargs):
        """ create task for submit/give up/rollback """
        return Task.objects.create(
            instance=self,
            node=self.cur_node,
            user=operator,
            **kwargs
        )

    def get_transitions(self, only_agree=False, only_can_auto_agree=False):
        qs = Transition.objects.filter(
            process=self.process, input_node=self.cur_node
        ).order_by('is_agree', 'oid', 'id')
        if only_agree:
            qs = qs.filter(is_agree=True)
        if only_can_auto_agree:
            qs = qs.filter(can_auto_agree=True)
        return [e for e in qs if e.is_match_condition(self.content_object)]

    def get_agree_transitions(self, only_can_auto_agree=True):
        return self.get_transitions(only_agree=True, only_can_auto_agree=only_can_auto_agree)

    def get_agree_transition(self, only_can_auto_agree=True):
        agree_transitions = self.get_agree_transitions(only_can_auto_agree)
        return agree_transitions[0] if agree_transitions else None

    def get_merged_agree_transitions(self):
        transitions = self.get_transitions(self)

        transition_names = [e.name for e in transitions]
        merged_transitions = []
        simple_agree_added = False

        for transition in transitions:
            if not transition.is_agree:
                merged_transitions.append(transition)
            elif len(transition_names) == 1:
                merged_transitions.append(transition)
            elif not simple_agree_added:  # use which transition depend on condition
                simple_agree_added = True
                merged_transitions.append(transition.as_simple_agree_transition())

        return merged_transitions

    def get_todo_task(self, user=None):
        return self.get_todo_tasks(user).first()

    def get_todo_tasks(self, user=None):
        qs = Task.objects.filter(
            instance=self, node=self.cur_node,
            status='in progress').filter(Q(agent_user=user) | Q(user=user))
        if user:
            qs = qs.filter(Q(agent_user=user) | Q(user=user))
        return qs

    def is_user_agreed(self, user):
        events = Event.objects.filter(instance=self).order_by('-created_on', '-id')
        users = []
        for event in events:
            if event.act_type in ['give up', 'reject', 'back to']:
                break
            if event.act_type != 'agree':
                continue
            users.append(event.user)
        return user in users

    def get_can_back_to_activities(self):
        events = Event.objects.filter(instance=self).order_by('-created_on', '-id')
        activities = []
        for event in events:
            if event.old_node == self.cur_node:
                activities = []
                continue
            if event.old_node.status not in ['in progress']:
                break
            if event.old_node not in activities:
                activities.append(event.old_node)
        return activities

    def has_received(self):
        if self.cur_node.status != 'in progress':
            return True
        return self.task_set.filter(status='in progress', receive_on__isnull=False).exists()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        wf_obj = self.content_object
        if not self.no and wf_obj:  # self.no depend on pk(you should save to get pk)
            self.no = wf_obj.get_process_no()
            self.created_by = wf_obj.created_by
            self.summary = wf_obj.get_process_summary()
            super().save(force_update=True)


class Authorization(models.Model):
    user = models.ForeignKey(
        AUTH_USER_MODEL, verbose_name='User',
        null=True, on_delete=models.SET_NULL,
        related_name='authorized_user_authorizations')
    agent_user = models.ForeignKey(
        AUTH_USER_MODEL, verbose_name='Agent user',
        null=True, on_delete=models.SET_NULL,
        related_name='agent_user_authorizations')
    processes = models.ManyToManyField(Process, verbose_name='Processes')
    start_on = models.DateField('Start on')
    end_on = models.DateField('End on')

    def __str__(self):
        return '%s %s %s %s' % (self.user, self.agent_user, self.start_on, self.end_on)

    def update_agent_for_task(self):
        Task.objects.filter(
            authorization=self, status='in progress'
        ).update(agent_user=None)
        Task.objects.filter(
            user=self.authorized_user, status='in progress',
            instance__process__in=self.processes.all(),
        ).update(agent_user=self.agent_user, authorization=self)

    def is_active(self):
        today = datetime.date.today()
        if self.start_on > today:
            return False
        if self.end_on < today:
            return False
        return True

    def save(self, *args, **kwargs):
        # TODO on delete
        super().save(*args, **kwargs)
        self.update_agent_for_task()


class Task(models.Model):
    """
    A task object represents a task you are performing.
    """
    STATUS_CHOICES = (
        ('in progress', 'In Progress'),
        ('completed', 'Completed'),
    )

    instance = models.ForeignKey(
        ProcessInstance,
        on_delete=models.CASCADE
    )
    node = models.ForeignKey(
        Node,
        on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        AUTH_USER_MODEL, verbose_name='User', null=True, blank=True,
        on_delete=models.SET_NULL
    )
    agent_user = models.ForeignKey(
        AUTH_USER_MODEL, verbose_name='Agent user',
        related_name='agent_user_tasks',
        null=True, blank=True,
        on_delete=models.SET_NULL
    )
    authorization = models.ForeignKey(
        Authorization, verbose_name='Authorization',
        blank=True, null=True, on_delete=models.SET_NULL,
    )
    status = models.CharField(max_length=255, choices=STATUS_CHOICES, default='in progress')
    receive_on = models.DateTimeField('Receive on', null=True, blank=True)
    is_hold = models.BooleanField('Is hold', default=False)
    is_joint = models.BooleanField('Is joint', default=False)

    created_on = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return '%s - %s - %s' % (self.instance.summary, self.node.name, self.pk)

    def update_authorization(self, commit=True):
        today = datetime.date.today()
        authorization = Authorization.objects.filter(
            processes=self.instance.process, user=self.user,
            start_on__lte=today, end_on__gte=today,
        ).first()

        self.agent_user = None
        if authorization:
            self.agent_user = authorization.agent_user

        if commit:
            self.save()


class Event(models.Model):
    """
    A task perform log.
    """
    EVENT_ACT_CHOICES = (
        ('transition', 'Transition'),
        ('agree', 'Agree'),
        ('edit', 'Edit'),
        ('give up', 'Give up'),
        ('reject', 'Reject'),
        ('back to', 'Back to'),
        ('rollback', 'Rollback'),
        ('comment', 'Comment'),
        ('assign', 'Assign'),  # assign another user to audit
        ('hold', 'Hold'),
        ('unhold', 'Unhold')
    )
    instance = models.ForeignKey(
        ProcessInstance,
        on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        on_delete=models.CASCADE
    )
    act_type = models.CharField(
        max_length=255, choices=EVENT_ACT_CHOICES,
        default='transition')
    act_name = models.CharField(
        max_length=255, blank=True)
    old_node = models.ForeignKey(
        Node, related_name='out_events',
        null=True, blank=True,
        on_delete=models.CASCADE
    )
    new_node = models.ForeignKey(
        Node, related_name='in_events',
        null=True, blank=True,
        on_delete=models.CASCADE
    )
    task = models.ForeignKey(
        Task, related_name='events',
        null=True, blank=True,
        on_delete=models.SET_NULL
    )

    next_operators = models.ManyToManyField(
        AUTH_USER_MODEL, related_name='audit_events',
        blank=True)
    notice_users = models.ManyToManyField(
        AUTH_USER_MODEL, related_name='notice_events',
        blank=True)

    comment = models.TextField(blank=True, default='')
    attachments = models.ManyToManyField(
        LBAttachment, verbose_name='Attachment', blank=True)
    ext_data = JSONField(null=True, blank=True)

    created_on = models.DateTimeField(auto_now_add=True)

    def get_act_name(self):
        if self.act_type == 'transition':
            return self.act_name
        # TODO assign should show assign to who
        return self.get_act_type_display()

    def get_next_notice_users_display(self):
        if self.old_node == self.new_node:
            return ''
        return ', '.join([GET_USER_DISPLAY_NAME_FUNC(e) for e in self.notice_users.all()])

    def __str__(self):
        old_node = self.old_node.name if self.old_node else ''
        new_node = self.new_node.name if self.new_node else ''
        return '%s: %s - %s - %s' % (self.instance.summary, old_node,
                                     self.get_act_name(), new_node)


class BaseWFObj(models.Model):
    """
    A abstract class for flow model. Every flow model should inherit from it.
    """
    pinstance = models.ForeignKey(
        ProcessInstance, blank=True, null=True,
        related_name="%(class)s",
        verbose_name='Process instance',
        on_delete=models.CASCADE
    )
    created_on = models.DateTimeField('Created on', auto_now_add=True)
    created_by = models.ForeignKey(
        AUTH_USER_MODEL,
        null=True, on_delete=models.SET_NULL,
        verbose_name='Created by')

    class Meta:
        abstract = True

    def get_process_no(self):
        instance = self.pinstance
        if instance and instance.pk:
            return '%s%s' % (instance.process.prefix, instance.pk)
        return ''

    def get_status(self):
        return self.pinstance.cur_node.status

    def get_process_summary(self):
        return "%s" % self

    def update_process_summary(self, commit=True):
        pinstance = self.pinstance
        if not pinstance:
            return
        pinstance.summary = self.get_process_summary()
        if commit:
            pinstance.save()

    def on_complete(self):
        """ Will call when process complete """
        pass

    def on_submit(self):
        """ Will call when process submit """
        pass

    def on_fail(self):
        """ Will call when process fail(cancel/give up/reject)"""
        pass

    def on_do_transition(self, cur_node, to_node):
        """
        Will call when process node transfer
        """
        pass

    def get_absolute_url(self):
        return ''

    def save(self, *args, **kwargs):
        """
        update self.pinstance.summary on save.
        """
        super().save(*args, **kwargs)
        instance = self.pinstance
        if instance:
            instance.summary = self.get_process_summary()
            instance.save()

    def create_pinstance(self, process, submit=False):
        """
        Create and set self.pinstance for this model

        :param process: Which process to use
        :param submit: Whether auto submit it after create
        :return: The created process instance
        """
        created_by = self.created_by
        if not isinstance(process, Process):
            process = get_or_none(Process, code=process)
        if not self.pk:
            self.save()
        instance = ProcessInstance.objects.create(
            process=process, created_by=created_by, content_object=self,
            cur_node=process.get_draft_active())
        self.pinstance = instance
        self.save()  # instance will save after self.save
        if submit:
            self.submit_process()
        return instance

    def submit_process(self, user=None):
        """
        Submit this process.

        :param user: Which user submit the process. The user is self.created_by if user is None.
        """
        from lbworkflow.core.transition import TransitionExecutor

        instance = self.pinstance
        if instance.cur_node.is_submitted():
            return
        user = user or self.created_by
        task = instance.create_task(user)
        transition = instance.get_transitions()[0]
        TransitionExecutor(user, instance, task, transition).execute()



from django.urls import reverse

from .test_base import BaseTests


class ViewTests(BaseTests):

    def test_my_wf(self):
        self.client.login(username='owner', password='password')
        resp = self.client.get(reverse('wf_my_wf'))
        self.assertEqual(resp.status_code, 200)

    def test_list_wf(self):
        self.client.login(username='owner', password='password')
        resp = self.client.get(reverse('wf_list_wf'))
        self.assertEqual(resp.status_code, 200)

    def test_todo(self):
        self.client.login(username='owner', password='password')
        resp = self.client.get(reverse('wf_todo'))
        self.assertEqual(resp.status_code, 200)



"""
Django settings for testproject project.

Generated by 'django-admin startproject' using Django 1.10.6.

For more information on this file, see
https://docs.djangoproject.com/en/1.10/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.10/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.10/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '*lx!g2o%m)b613n$709334=+ulwi^&6e8=o6h3upwn4&3c$o^p'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'crispy_forms',
    'lbattachment',
    'lbadminlte',
    'lbutils',
    'compressor',
    'djangobower',
    'django_select2',
    'bootstrap_pagination',

    'lbworkflow',
    'lbworkflow.simplewf',
    'lbworkflow.tests.leave',
    'lbworkflow.tests.purchase',

    'lbworkflow.tests.issue',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'lbworkflow.tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'testproject.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.10/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}


# Password validation
# https://docs.djangoproject.com/en/1.10/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/1.10/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.10/howto/static-files/

STATIC_URL = '/static/'

LBWF_APPS = {
    'leave': 'lbworkflow.tests.leave',
    'purchase': 'lbworkflow.tests.purchase',
    'simplewf': 'lbworkflow.simplewf',
}

STATIC_ROOT = os.path.join(BASE_DIR, 'collectedstatic')

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

CRISPY_TEMPLATE_PACK = 'bootstrap3'

# bower
STATICFILES_FINDERS += (('djangobower.finders.BowerFinder'),)
BOWER_COMPONENTS_ROOT = BASE_DIR

BOWER_INSTALLED_APPS = (
    'admin-lte#2.3.11',
    'font-awesome#4.7.0',
    'ionicons#2.0.1',

    'modernizr',
    # POLYFILLS: javascript fallback solutions for older browsers.
    # CSS3 selectors for IE 6-8.
    'selectivizr',
    # min/max width media queries for IE 6-8.
    'respond',
    # CSS3 styles for IE 6-8.
    'pie',
    # HTML5 tag support for IE 6-8.
    'html5shiv',

    'masonry#4.1.1',
    'blueimp-file-upload#9.12.5',
    'flatpickr-calendar#2.5.6',
)

# django-compressor
STATICFILES_FINDERS += (('compressor.finders.CompressorFinder'),)
COMPRESS_PRECOMPILERS = (
    ('text/coffeescript', 'coffee --compile --stdio'),
    ('text/less', 'lessc {infile} {outfile}'),
    ('text/x-sass', 'sass {infile} {outfile}'),
    ('text/x-scss', 'sass --scss {infile} {outfile}'),
)

PROJECT_TITLE = 'LB-Workflow'



from django.contrib import admin

from .models import SimpleWorkFlow


class SimpleWorkFlowAdmin(admin.ModelAdmin):
    list_display = ('summary', 'content')


admin.site.register(SimpleWorkFlow, SimpleWorkFlowAdmin)



import importlib

from django.contrib import messages
from django.db.models import Q
from django.utils import timezone

from lbworkflow import settings


def import_wf_views(wf_code, view_module_name='views'):
    wf_module = settings.WF_APPS.get(wf_code.split('__')[0])
    return importlib.import_module('%s.%s' % (wf_module, view_module_name))


def add_processed_message(request, process_instance, act_descn='Processed'):
    messages.info(
        request,
        'Process "%s" has been %s. Current status："%s" Current user："%s"' %
        (
            process_instance.no, act_descn, process_instance.cur_node.name,
            process_instance.get_operators_display()
        )
    )


def get_wf_template_names(wf_code, base_template_name, wf_object=None, model=None):
    templates = []
    paths = wf_code.split('__')
    for i in range(len(paths)):
        temp_paths = paths[:len(paths) - i] + [base_template_name]
        templates.append('/'.join(temp_paths))
    _meta = None
    if wf_object:
        _meta = wf_object._meta
    elif model:
        _meta = model._meta
    if _meta:
        app_label = _meta.app_label
        object_name = _meta.object_name.lower()
        templates.extend([
            "%s/%s/%s" % (app_label, object_name, base_template_name,),
            "%s/%s" % (app_label, base_template_name,), ])
    return templates


def user_wf_info_as_dict(wf_obj, user):
    ctx = {}
    if user.is_anonymous:
        return ctx
    instance = wf_obj.pinstance
    is_wf_admin = instance.is_wf_admin(user)
    in_process = instance.cur_node.status == 'in progress'
    task = instance.get_todo_task(user)
    ctx['wf_code'] = instance.process.code
    ctx['process'] = instance.process
    ctx['process_instance'] = instance
    ctx['object'] = wf_obj
    ctx['task'] = task
    ctx['wf_history'] = instance.event_set.all().order_by('-created_on', '-pk')
    ctx['operators_display'] = instance.get_operators_display()
    ctx['is_wf_admin'] = is_wf_admin

    can_edit = not instance.cur_node.is_submitted() and instance.created_by == user
    can_edit = can_edit or (instance.cur_node.can_edit and task)
    can_edit = can_edit or is_wf_admin
    ctx['can_edit'] = can_edit
    ctx['can_rollback'] = instance.can_rollback(user)

    if in_process:
        ctx['can_assign'] = task or is_wf_admin or user.is_superuser
        ctx['can_remind'] = instance.created_by == user or is_wf_admin
        ctx['can_give_up'] = instance.can_give_up(user)

    if task:
        instance.get_todo_tasks(user).filter(
            receive_on=None
        ).update(receive_on=timezone.now())
        transitions = instance.get_transitions()
        ctx['can_reject'] = instance.cur_node.can_reject
        ctx['can_back_to'] = None
        ctx['transitions'] = transitions
        ctx['agree_transitions'] = instance.get_merged_agree_transitions()
        ctx['other_transitions'] = [e for e in transitions if not e.is_agree]
    # TODO add reject,given up to other_transitions?
    return ctx


def get_base_wf_permit_query_param(user, process_instance_field_prefix='pinstance__'):
    def p(param_name, value):
        return {process_instance_field_prefix + param_name: value}
    q_param = Q()
    # Submit
    q_param = q_param | Q(
        **p('created_by', user)
    )
    # share
    q_param = q_param | Q(
        **p('can_view_users', user)
    )
    # Can process
    q_param = q_param | Q(
        **p('task__user', user)
    )
    q_param = q_param | Q(
        **p('task__agent_user', user)
    )
    return q_param



from django.db import models

from lbworkflow.models import BaseWFObj


class Purchase(BaseWFObj):
    title = models.CharField('Title', max_length=255)
    reason = models.CharField('Reason', max_length=255)

    def __str__(self):
        return self.reason


class Item(models.Model):
    purchase = models.ForeignKey(
        Purchase,
        on_delete=models.CASCADE,
    )
    name = models.CharField('Name', max_length=255)
    qty = models.IntegerField('Qty')
    note = models.CharField('Note', max_length=255)

    class Meta:
        verbose_name = 'Purchase Item'

    def __str__(self):
        return self.name



from django.conf import settings as django_settings

AUTH_USER_MODEL = getattr(django_settings, 'AUTH_USER_MODEL', 'auth.User')

USER_PARSER = getattr(
    django_settings, 'LBWF_USER_PARSER',
    'lbworkflow.core.userparser.SimpleUserParser')

WF_PAGE_SIZE = getattr(django_settings, 'LBWF_PAGE_SIZE', 20)

EVAL_FUNCS = getattr(django_settings, 'LBWF_EVAL_FUNCS', {})

WF_SEND_MSG_FUNCS = getattr(
    django_settings,
    'LBWF_WF_SEND_MSG_FUNCS',
    ['lbworkflow.core.sendmsg.wf_print', ],
)

GET_USER_DISPLAY_NAME_FUNC = getattr(
    django_settings, 'LBWF_GET_USER_DISPLAY_NAME_FUNC',
    lambda user: "%s" % user)

CAN_EDIT_WF_FUNC = getattr(
    django_settings, 'LBWF_CAN_EDIT_WF_FUNC',
    lambda *args, **kwargs: True)

CAN_SUBMIT_WF_FUNC = getattr(
    django_settings, 'LBWF_CAN_SUBMIT_WF_FUNC',
    lambda *args, **kwargs: True)

CAN_VIEW_WF_FUNC = getattr(
    django_settings, 'LBWF_CAN_VIEW_WF_FUNC',
    lambda *args, **kwargs: True)

DEBUG_WORKFLOW = getattr(django_settings, 'LBWF_DEBUG_WORKFLOW', False)
WF_APPS = getattr(django_settings, 'LBWF_APPS', {})

PROCESS_INSTANCE_GET_PERMIT_QUERY_PARAM_FUNC = getattr(
    django_settings, 'LBWF_PROCESS_INSTANCE_GET_PERMIT_QUERY_PARAM_FUNC',
    lambda user, query_param: query_param)

QUICK_SEARCH_FORM = getattr(
    django_settings, 'LBWF_QUICK_SEARCH_FORM',
    'lbworkflow.forms.BSQuickSearchForm')

QUICK_SEARCH_WITH_EXPORT_FORM = getattr(
    django_settings, 'LBWF_QUICK_SEARCH_WITH_EXPORT_FORM',
    'lbworkflow.forms.BSQuickSearchWithExportForm')

WORK_FLOW_FORM = getattr(
    django_settings, 'LBWF_WORK_FLOW_FORM',
    'lbworkflow.forms.BSWorkFlowForm')

BATCH_WORK_FLOW_FORM = getattr(
    django_settings, 'LBWF_BATCH_WORK_FLOW_FORM',
    'lbworkflow.forms.BSBatchWorkFlowForm')

BACK_TO_ACTIVITY_FORM = getattr(
    django_settings, 'LBWF_BACK_TO_ACTIVITY_FORM',
    'lbworkflow.forms.BSBackToNodeForm')

ADD_ASSIGNEE_FORM = getattr(
    django_settings, 'LBWF_ADD_ASSIGNEE_FORM',
    'lbworkflow.forms.BSAddAssigneeForm')



from crispy_forms.bootstrap import StrictButton
from crispy_forms.layout import Layout
from django import forms
from lbutils import BootstrapFormHelperMixin

from lbworkflow.forms import BSQuickSearchForm
from lbworkflow.forms import WorkflowFormMixin

from .models import SimpleWorkFlow


class SearchForm(BSQuickSearchForm):
    def layout(self):
        self.helper.layout = Layout(
            'q_quick_search_kw',
            StrictButton('Search', type="submit", css_class='btn-sm btn-default'),
            StrictButton('Export', type="submit", name="export", css_class='btn-sm btn-default'),
        )


class SimpleWorkFlowForm(BootstrapFormHelperMixin, WorkflowFormMixin, forms.ModelForm):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.init_crispy_helper()
        self.layout_fields([
            ['summary', ],
            ['content', ],
        ])

    class Meta:
        model = SimpleWorkFlow
        fields = [
            'summary', 'content'
        ]



from lbworkflow.core.datahelper import create_app


def load_data():
    load_base()


def load_base():
    create_app('5f31d065-00aa-0010-beea-641f0a670010', 'Simple', action='wf_execute_transition')
    create_app('5f31d065-00aa-0010-beea-641f0a670020', 'Customized URL')



from lbworkflow.flowgen import FlowAppGenerator
from lbworkflow.flowgen import clean_generated_files
from lbworkflow.tests.issue.models import Issue
from lbworkflow.tests.purchase.models import Item
from lbworkflow.tests.purchase.models import Purchase

from .test_base import BaseTests


class ViewTests(BaseTests):

    def test_gen_no_item_list(self):
        FlowAppGenerator().gen(Issue)
        clean_generated_files(Issue)

    def test_gen_with_item_list(self):
        FlowAppGenerator().gen(Purchase, [Item])
        clean_generated_files(Purchase)



from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q
from django.forms import ModelForm
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.views.generic.base import TemplateResponseMixin
from django.views.generic.edit import FormView
from lbutils import as_callable
from lbutils import get_or_none

from lbworkflow import settings
from lbworkflow.core.exceptions import HttpResponseException
from lbworkflow.core.sendmsg import wf_send_msg
from lbworkflow.core.transition import TransitionExecutor
from lbworkflow.core.transition import create_event
from lbworkflow.models import Node
from lbworkflow.models import ProcessInstance
from lbworkflow.models import Task
from lbworkflow.models import Transition
from lbworkflow.settings import GET_USER_DISPLAY_NAME_FUNC

from .forms import FormsView
from .helper import add_processed_message
from .helper import import_wf_views
from .helper import user_wf_info_as_dict


class ExecuteTransitionView(TemplateResponseMixin, FormsView):
    """
    request.GET:
        ts_id: transition pk
        wi_id: work item pk
        pid: process instance pk (not used)
    """
    form_classes = {
        'form': as_callable(settings.WORK_FLOW_FORM)
    }

    def get_success_url(self):
        return reverse("wf_todo")

    def get_template_names(self):
        try:
            return super().get_template_names()
        except ImproperlyConfigured:
            base_tmpl = 'lbworkflow/do_transition_form.html'
            _meta = self.object._meta
            app_label = _meta.app_label
            object_name = _meta.object_name.lower()
            return ["%s/%s/%s" % (app_label, object_name, base_tmpl,),
                    "%s/%s" % (app_label, base_tmpl,),
                    base_tmpl]

    def get_init_transition(self, process_instance, request):
        ts_id = request.GET.get('ts_id')
        return Transition.objects.get(pk=ts_id)

    def get_task(self, request):
        # TODO admin may don't have task, need auto create a work item for admin
        wi_id = request.GET.get('wi_id')
        user = request.user
        return get_or_none(Task, Q(user=user) | Q(agent_user=user), id=wi_id)

    def init_process_data(self, request):
        task = self.get_task(request)
        if not task:
            self.raise_no_permission_exception()
        instance = task.instance

        self.transition = self.get_init_transition(instance, request)
        self.task = task
        self.process_instance = instance
        self.object = instance.content_object

    def raise_no_permission_exception(self, instance=None):
        from django.template import Context, Template
        t = Template("""
            <!DOCTYPE HTML>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <title></title>
            </head>
            <body>
                No permission to perform this action
                {% if instance %}
                    <br/>
                    <a href="{% url 'wf_detail' instance.pk %}"> View this process </a>
                {% endif %}
            </body>
            </html>
            """)
        http_response = HttpResponse(
            t.render(Context({"instance": instance})), content_type='text/html', status=403)
        raise HttpResponseException(http_response)

    def has_permission(self, request, instance, task, transition):
        node_is_ok = (
            transition.input_node == instance.cur_node
            and task.node == instance.cur_node
            and task.status == 'in progress')
        user_is_ok = (
            request.user in [task.user, task.agent_user]
            or instance.is_wf_admin(request.user))
        is_ok = node_is_ok and user_is_ok
        return is_ok

    def check_permission(self, request):
        """ If no permission raise HttpResponseException """
        if not self.has_permission(request, self.process_instance, self.task, self.transition):
            self.raise_no_permission_exception(self.process_instance)

    def get_transition_before_execute(self, cleaned_data):
        return self.transition

    def do_transition(self, cleaned_data):
        comment = cleaned_data.get('comment')
        attachments = cleaned_data.get('attachments')

        user = self.request.user
        instance = self.process_instance
        transition = self.get_transition_before_execute(cleaned_data)

        TransitionExecutor(user, instance, self.task, transition, comment, attachments).execute()

    def add_processed_message(self, process_instance, act_descn='Processed'):
        add_processed_message(self.request, process_instance, act_descn)

    def save_form(self, form):
        return form.save()

    def forms_valid(self, **forms):
        form = forms.pop('form')
        if isinstance(form, ModelForm):
            wf_obj = self.save_form(form)
            # update cache for wf_obj
            self.object = wf_obj
            self.process_instance = wf_obj.pinstance
        for other_form in forms:
            self.save_form(other_form)
        self.do_transition(form.cleaned_data)
        self.add_processed_message(self.process_instance)
        return HttpResponseRedirect(self.get_success_url())

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['task'] = self.task
        kwargs['transition'] = self.transition
        kwargs.update(user_wf_info_as_dict(self.object, self.request.user))
        return kwargs

    def dispatch(self, request, *args, **kwargs):
        self.request = request
        try:
            self.init_process_data(request)
            self.check_permission(request)
            return super().dispatch(request, *args, **kwargs)
        except HttpResponseException as error:
            return error.http_response


class BatchExecuteTransitionView(FormView):
    template_name = 'lbworkflow/batch_transition_form.html'
    form_class = as_callable(settings.BATCH_WORK_FLOW_FORM)

    def get_success_url(self):
        return reverse("wf_todo")

    def get_context_data(self, **kwargs):
        kwargs['task_list'] = self.task_list
        kwargs['transition_name'] = self.get_transition_name()
        return super().get_context_data(**kwargs)

    def get_transition_name(self):
        return 'Agree'

    def get_transition(self, process_instance):
        pass

    def add_processed_message(self, process_instance, act_descn='Processed'):
        add_processed_message(self.request, process_instance, act_descn)

    def post(self, request, *args, **kwargs):
        if not request.POST.get('do_submit'):
            return self.get(request, *args, **kwargs)
        return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        user = self.request.user
        cleaned_data = form.cleaned_data
        for task in self.task_list:
            if task.user != user:
                # TODO message for ignore
                continue
            instance = task.instance
            transition = self.get_transition(task.instance)
            comment = cleaned_data.get('comment')
            attachments = cleaned_data.get('attachments')
            TransitionExecutor(
                user, instance, task, transition, comment, attachments
            ).execute()
            self.add_processed_message(instance)
        return HttpResponseRedirect(self.get_success_url())

    def get_task_list(self, request):
        task_pk_list = request.POST.getlist('wi')
        task_pk_list = [e for e in task_pk_list if e]
        task_list = Task.objects.filter(status='in progress', pk__in=task_pk_list)
        return task_list

    def dispatch(self, request, *args, **kwargs):
        self.task_list = self.get_task_list(request)
        return super().dispatch(request, *args, **kwargs)


class ExecuteBackToTransitionView(ExecuteTransitionView):
    form_classes = {
        'form': as_callable(settings.BACK_TO_ACTIVITY_FORM)
    }

    def has_permission(self, request, instance, task, transition):
        # TODO ...
        return True

    def get_form_kwargs(self, form_class_key, form_class):
        """
        Returns the keyword arguments for instantiating the form.
        """
        kwargs = super().get_form_kwargs(form_class_key, form_class)
        kwargs['process_instance'] = self.process_instance
        return kwargs

    def get_init_transition(self, process_instance, request):
        return process_instance.get_back_to_transition()

    def get_transition_before_execute(self, cleaned_data):
        back_to_node = cleaned_data.get('back_to_node')
        transition = self.transition
        transition.output_node = Node.objects.get(pk=back_to_node)
        return transition


class ExecuteGiveUpTransitionView(ExecuteTransitionView):

    def get_success_url(self):
        return reverse("wf_my_wf")

    def get_task(self, request):
        pk = request.GET.get('pk')
        process_instance = get_or_none(ProcessInstance, pk=pk)
        if not process_instance:
            return None
        return process_instance.create_task(request.user)
        return Task(
            instance=process_instance,
            node=process_instance.cur_node,
            user=request.user)

    def has_permission(self, request, instance, task, transition):
        return instance.can_give_up(request.user)

    def get_init_transition(self, process_instance, request):
        return process_instance.get_give_up_transition()


class BatchExecuteGiveUpTransitionView(BatchExecuteTransitionView):

    def get_success_url(self):
        return reverse("wf_my_wf")

    def get_task_list(self, request):
        instance_pk_list = request.POST.getlist('pi')
        instance_list = ProcessInstance.objects.filter(
            cur_node__status='in progress',
            pk__in=instance_pk_list)
        task_list = []
        for instance in instance_list:
            task = Task(
                instance=instance,
                node=instance.cur_node,
                user=instance.created_by,
            )
            task_list.append(task)
        return task_list

    def get_transition_name(self):
        return 'Give up'

    def get_transition(self, process_instance):
        return process_instance.get_give_up_transition()


class ExecuteAgreeTransitionView(ExecuteTransitionView):
    def get_init_transition(self, process_instance, request):
        return process_instance.get_agree_transition(False)

    def get_transition_before_execute(self, cleaned_data):
        return self.process_instance.get_agree_transition(False)


class BatchExecuteAgreeTransitionView(BatchExecuteTransitionView):
    def get_transition_name(self):
        return 'Agree'

    def get_transition(self, process_instance):
        return process_instance.get_agree_transition(False)


class ExecuteRejectTransitionView(ExecuteTransitionView):
    def get_task(self, request):
        task = super().get_task(request)
        return task if task and task.node.can_reject else None

    def get_init_transition(self, process_instance, request):
        return process_instance.get_reject_transition()


class BatchExecuteRejectTransitionView(BatchExecuteTransitionView):
    def get_task_list(self, request):
        task_list = []
        for task in super().get_task_list(request):
            if task.node.can_reject:
                task_list.append(task)
            else:
                messages.info(
                    request,
                    """You can't reject process "%s" """ %
                    (
                        task.instance.no,
                    )
                )
        return task_list

    def get_transition_name(self):
        return 'Reject'

    def get_transition(self, process_instance):
        return process_instance.get_reject_transition()


# TODO Rollback


class AddAssigneeView(ExecuteTransitionView):
    form_classes = {
        'form': as_callable(settings.ADD_ASSIGNEE_FORM)
    }

    def get_form_kwargs(self, form_class_key, form_class):
        kwargs = super().get_form_kwargs(form_class_key, form_class)
        kwargs['instance'] = self.process_instance
        return kwargs

    def get_init_transition(self, process_instance, request):
        return process_instance.get_add_assignee_transition()

    def do_transition(self, cleaned_data):
        request = self.request
        transition = self.get_transition_before_execute(cleaned_data)
        comment = cleaned_data.get('comment')
        attachments = cleaned_data.get('attachments')
        user = request.user
        instance = self.process_instance
        assignees = cleaned_data.get('assignees', [])
        for assignee in assignees:
            instance.create_task(assignee, is_joint=True)
        msg = 'Add assignee %s.' % ', '.join([GET_USER_DISPLAY_NAME_FUNC(e) for e in assignees])
        messages.info(request, msg)
        comment = msg + '\n' + comment
        event = create_event(
            instance, transition,
            comment=comment, user=user,
            old_node=instance.cur_node, new_node=instance.cur_node)
        event.attachments.add(*attachments)
        wf_send_msg(assignees, 'new_workitem', event)


def execute_transitions(request, wf_code, trans_func):
    views = import_wf_views(wf_code, 'wf_views')
    # TODO check permission
    # TODO get transition and check if this transition can call this function
    func = getattr(views, trans_func)
    return func(request)



from django.urls import reverse

from .test_base import BaseTests


class ViewTests(BaseTests):

    def test_flowchart(self):
        resp = self.client.get(reverse('wf_process_flowchart', args=('leave', )))
        self.assertEqual(resp.status_code, 200)

    def test_process_instance_flowchart(self):
        leave = self.create_leave('flowchart', True)
        resp = self.client.get(
            reverse('wf_process_instance_flowchart',
                    args=(leave.pinstance.pk, ))
        )
        self.assertEqual(resp.status_code, 200)



from django.contrib.auth import get_user_model
from django.urls import reverse

from lbworkflow.core.transition import TransitionExecutor
from lbworkflow.views.helper import user_wf_info_as_dict

from .leave.models import Leave
from .test_base import BaseTests

User = get_user_model()


class TransitionExecutorTests(BaseTests):

    def test_submit(self):
        leave = self.leave
        instance = self.leave.pinstance
        leave.submit_process()

        # A1 will auto agree
        self.assertEqual(leave.pinstance.cur_node.name, 'A2')
        self.assertEqual(leave.pinstance.get_operators_display(), 'tom')

        # A3 not auto agree
        task = instance.get_todo_task()
        transition = instance.get_agree_transition()
        TransitionExecutor(self.users['tom'], instance, task, transition).execute()
        self.assertEqual(leave.pinstance.cur_node.name, 'A3')


class ViewTests(BaseTests):

    def get_transition_url(self, leave, user):
        ctx = user_wf_info_as_dict(leave, user)

        transitions = ctx['transitions']
        transition = transitions[0]
        transition_url = transition.get_app_url(ctx['task'])
        return transition_url

    def setUp(self):
        super(ViewTests, self).setUp()
        self.leave.submit_process()

        leave = self.leave
        ctx = user_wf_info_as_dict(leave, self.users['tom'])

        transitions = ctx['transitions']
        transition = transitions[0]
        self.transition_url = transition.get_app_url(ctx['task'])

        self.task = ctx['task']

        self.client.login(username='tom', password='password')

    def do_agree(self, username, node_name, leave=None, data={}):
        if not leave:
            leave = self.leave
        leave = Leave.objects.get(pk=leave.pk)

        self.client.login(username=username, password='password')

        transition_url = self.get_transition_url(leave, self.users[username])
        resp = self.client.post(transition_url, data=data)
        self.assertRedirects(resp, '/wf/todo/')
        leave = Leave.objects.get(pk=leave.pk)
        self.assertEqual(node_name, leave.pinstance.cur_node.name)

    def test_execute_transition(self):
        self.do_agree('tom', 'A3')
        self.do_agree('vicalloy', 'A4')

    def test_execute_transition_customized_url(self):
        self.do_agree('tom', 'A3')
        self.do_agree('vicalloy', 'A4')
        data = {
            'actual_start_on': '2017-04-25 08:00',
            'actual_end_on': '2017-04-26 08:00',
            'actual_leave_days': '2',
        }
        self.do_agree('hr', 'Completed', data=data)

    def goto_A2B1(self):
        leave = self.leave
        leave.leave_days = 10
        leave.save()

        self.do_agree('tom', 'A2B1')

    def test_execute_transition_with_condition(self):
        self.goto_A2B1()

    def test_execute_transition_joint(self):
        self.goto_A2B1()
        self.do_agree('tom', 'A2B1')
        self.do_agree('owner', 'A3')

    def test_execute_transition_no_permission(self):
        self.client.login(username='vicalloy', password='password')
        resp = self.client.post(self.transition_url)
        self.assertEqual(resp.status_code, 403)

    def test_simple_agree(self):
        url = reverse('wf_agree')
        resp = self.client.get('%s?wi_id=%s' % (url, self.task.pk))
        self.assertEqual(resp.status_code, 200)

        resp = self.client.post('%s?wi_id=%s' % (url, self.task.pk))
        self.assertRedirects(resp, '/wf/todo/')
        leave = Leave.objects.get(pk=self.leave.pk)
        self.assertEqual('A3', leave.pinstance.cur_node.name)

    def test_reject(self):
        url = reverse('wf_reject')
        resp = self.client.get('%s?wi_id=%s' % (url, self.task.pk))
        self.assertEqual(resp.status_code, 200)

        resp = self.client.post('%s?wi_id=%s' % (url, self.task.pk))
        self.assertRedirects(resp, '/wf/todo/')
        leave = Leave.objects.get(pk=self.leave.pk)
        self.assertEqual('Rejected', leave.pinstance.cur_node.name)

    def test_give_up(self):
        self.client.login(username='owner', password='password')
        url = reverse('wf_give_up')
        resp = self.client.get('%s?pk=%s' % (url, self.leave.pinstance.pk))
        self.assertEqual(resp.status_code, 200)

        resp = self.client.post('%s?pk=%s' % (url, self.leave.pinstance.pk))
        self.assertRedirects(resp, '/wf/my/')
        leave = Leave.objects.get(pk=self.leave.pk)
        self.assertEqual('Given up', leave.pinstance.cur_node.name)

    def test_back_to(self):
        self.client.post(self.transition_url)  # A2 TO A3
        leave = Leave.objects.get(pk=self.leave.pk)

        url = reverse('wf_back_to')
        resp = self.client.get('%s?wi_id=%s' % (url, self.task.pk))
        self.assertEqual(resp.status_code, 200)
        back_to_node = leave.pinstance.get_can_back_to_activities()[0]

        resp = self.client.post(
            '%s?wi_id=%s' % (url, self.task.pk),
            {'back_to_node': back_to_node.pk}
        )
        self.assertRedirects(resp, '/wf/todo/')
        leave = Leave.objects.get(pk=self.leave.pk)
        self.assertEqual('A2', leave.pinstance.cur_node.name)

    def test_batch_agree(self):
        url = reverse('wf_batch_agree')
        ctx = user_wf_info_as_dict(self.leave, self.users['tom'])
        data = {
            'wi': [ctx['task'].pk, 1, 2, 3],
        }
        resp = self.client.post(url, data)
        self.assertEqual(resp.status_code, 200)

        data['do_submit'] = 'submit'
        resp = self.client.post(url, data)
        self.assertRedirects(resp, '/wf/todo/')
        leave = Leave.objects.get(pk=self.leave.pk)
        self.assertEqual('A3', leave.pinstance.cur_node.name)

    def test_batch_reject(self):
        url = reverse('wf_batch_reject')
        ctx = user_wf_info_as_dict(self.leave, self.users['tom'])
        data = {
            'wi': [ctx['task'].pk, 1, 2, 3],
        }
        resp = self.client.post(url, data)
        self.assertEqual(resp.status_code, 200)

        data['do_submit'] = 'submit'
        resp = self.client.post(url, data)
        self.assertRedirects(resp, '/wf/todo/')
        leave = Leave.objects.get(pk=self.leave.pk)
        self.assertEqual('Rejected', leave.pinstance.cur_node.name)

    def test_batch_give_up(self):
        self.client.login(username='owner', password='password')
        url = reverse('wf_batch_give_up')
        data = {
            'pi': [self.leave.pinstance.pk, 1, 2, 3],
        }
        resp = self.client.post(url, data)
        self.assertEqual(resp.status_code, 200)

        data['do_submit'] = 'submit'
        resp = self.client.post(url, data)
        self.assertRedirects(resp, '/wf/my/')
        leave = Leave.objects.get(pk=self.leave.pk)
        self.assertEqual('Given up', leave.pinstance.cur_node.name)

    def test_add_assignee(self):
        url = reverse('wf_add_assignee')
        url = '%s?wi_id=%s' % (url, self.task.pk)
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)

        data = {
            'assignees': (self.users['hr'].pk, self.users['owner'].pk, ),
            'comment': 'comments',
        }
        resp = self.client.post(url, data)
        self.assertRedirects(resp, '/wf/todo/')
        leave = Leave.objects.get(pk=self.leave.pk)
        self.assertEqual('A2', leave.pinstance.cur_node.name)

        self.do_agree('tom', 'A2')
        self.do_agree('hr', 'A2')
        self.do_agree('owner', 'A3')



from django.contrib.auth import get_user_model
from django.urls import reverse

from lbworkflow.core.datahelper import load_wf_data
from lbworkflow.simplewf.models import SimpleWorkFlow

from .test_base import BaseTests

User = get_user_model()


class SimpleWFTests(BaseTests):

    def create_wf(self, wf_code, summary, submit=True):
        wf = SimpleWorkFlow.objects.create(
            summary=summary,
            created_by=self.users['owner'])
        wf.create_pinstance(wf_code, submit)
        return wf

    def init_data(self):
        super().init_data()
        load_wf_data('lbworkflow.simplewf')
        self.wf_a_1 = self.create_wf('simplewf__A', 'wf_a_1')
        self.wf_a_2 = self.create_wf('simplewf__A', 'wf_a_2')
        self.wf_b_1 = self.create_wf('simplewf__B', 'wf_b_1')

    def setUp(self):
        super().setUp()
        self.client.login(username='owner', password='password')

    def test_wf_list(self):
        resp = self.client.get(reverse('wf_list', args=('simplewf__A', )))
        self.assertEqual(resp.status_code, 200)
        self.assertContains(resp, 'wf_a_1')
        self.assertContains(resp, 'wf_a_2')
        self.assertNotContains(resp, 'wf_b_1')

        resp = self.client.get(reverse('wf_list', args=('simplewf__B', )))
        self.assertEqual(resp.status_code, 200)
        self.assertNotContains(resp, 'wf_a_1')
        self.assertNotContains(resp, 'wf_a_2')
        self.assertContains(resp, 'wf_b_1')

    def test_wf_list_export(self):
        resp = self.client.get(reverse('wf_list', args=('simplewf__A', )), {'export': 1})
        self.assertEqual(resp.status_code, 200)

    def test_detail(self):
        resp = self.client.get(reverse('wf_detail', args=(self.wf_a_1.pinstance.pk, )))
        self.assertEqual(resp.status_code, 200)

    def test_submit(self):
        self.client.login(username='owner', password='password')

        url = reverse('wf_new', args=('simplewf__A', ))
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)

    def test_edit(self):
        self.client.login(username='admin', password='password')

        url = reverse('wf_edit', args=(self.wf_a_1.pinstance.pk, ))
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)



from django.contrib.auth import get_user_model
from lbutils import as_callable

from lbworkflow.models import App
from lbworkflow.models import Node
from lbworkflow.models import Process
from lbworkflow.models import ProcessCategory
from lbworkflow.models import Transition

User = get_user_model()


def get_or_create(cls, uid, **kwargs):
    uid_field_name = kwargs.pop('uid_field_name', 'uuid')
    obj = cls.objects.filter(**{uid_field_name: uid}).first()
    if obj:
        for k, v in kwargs.items():
            setattr(obj, k, v)
        obj.save()
        return obj
    kwargs[uid_field_name] = uid
    return cls.objects.create(**kwargs)


def create_user(username, **kwargs):
    password = kwargs.pop('password', 'password')
    user = User.objects.filter(username=username).first()
    if user:
        user.set_password(password)
        return user
    return User.objects.create_user(username, "%s@v.cn" % username, password, **kwargs)


def create_app(uuid, name, **kwargs):
    return get_or_create(App, uuid, name=name, **kwargs)


def create_category(uuid, name, **kwargs):
    return get_or_create(ProcessCategory, uuid, name=name, **kwargs)


def create_process(code, name, **kwargs):
    return get_or_create(Process, code, name=name, uid_field_name='code', **kwargs)


def create_node(uuid, process, name, **kwargs):
    return get_or_create(Node, uuid, process=process, name=name, **kwargs)


def get_node(process, name):
    """
    get node
    :param process:
    :param name: 'submit' or 'submit,5f31d065-4a87-487b-beea-641f0a6720c3'
    :return: node
    """
    name_and_uuid = [e.strip() for e in name.split(',') if e.strip()]
    qs = Node.objects.filter(process=process)
    if len(name_and_uuid) == 1:
        qs = qs.filter(name=name_and_uuid[0])
    else:
        qs = qs.filter(uuid=name_and_uuid[1])
    return qs[0]


def get_app(name):
    """
    get node
    :param process:
    :param name: 'submit' or 'submit,5f31d065-4a87-487b-beea-641f0a6720c3'
    :return: node
    """
    name_and_uuid = [e.strip() for e in name.split(',') if e.strip()]
    qs = App.objects
    if len(name_and_uuid) == 1:
        qs = qs.filter(name=name_and_uuid[0])
    else:
        qs = qs.filter(uuid=name_and_uuid[1])
    return qs[0]


def create_transition(uuid, process, from_node, to_node, app='Simple', **kwargs):
    from_node = get_node(process, from_node)
    to_node = get_node(process, to_node)
    app = get_app(app)
    return get_or_create(
        Transition, uuid, process=process, input_node=from_node,
        output_node=to_node, app=app, **kwargs)


def load_wf_data(app, wf_code=''):
    if wf_code:
        func = "%s.wfdata.load_%s" % (app, wf_code)
    else:
        func = "%s.wfdata.load_data" % app
    as_callable(func)()



from lbworkflow.views.transition import ExecuteTransitionView

from .forms import HRForm


class CustomizedTransitionView(ExecuteTransitionView):
    form_classes = {
        'form': HRForm
    }


c = CustomizedTransitionView.as_view()



from lbworkflow.views.generics import CreateView
from lbworkflow.views.generics import UpdateView
from lbworkflow.views.generics import WFListView

from .forms import LeaveForm
from .models import Leave


class LeaveCreateView(CreateView):
    form_classes = {
        'form': LeaveForm,
    }


new = LeaveCreateView.as_view()


class LeaveUpdateView(UpdateView):
    form_classes = {
        'form': LeaveForm,
    }


edit = LeaveUpdateView.as_view()


class LeaveListView(WFListView):
    wf_code = 'leave'
    model = Leave
    excel_file_name = 'leave'
    excel_titles = [
        'Created on', 'Created by',
        'Start on', 'End on', 'Leave days',
        'Actual start on', 'Actual start on', 'Actual leave days',
        'Status',
    ]

    def get_excel_data(self, o):
        return [
            o.created_by.username, o.created_on,
            o.start_on, o.end_on, o.leave_days,
            o.actual_start_on, o.actual_end_on, o.actual_leave_days,
            o.pinstance.cur_node.name,
        ]


show_list = LeaveListView.as_view()


def detail(request, instance, ext_ctx, *args, **kwargs):
    return {'for_test': 'from leave.detail'}



from lbworkflow.core.datahelper import create_user


def load_data():
    init_users()


def init_users():
    users = {
        'owner': create_user('owner'),
        'operator': create_user('operator'),
        'vicalloy': create_user('vicalloy'),
        'tom': create_user('tom'),
        'hr': create_user('hr'),
        'admin': create_user('admin', is_superuser=True, is_staff=True),
    }
    return users



# -*- coding: utf-8 -*-
from django import forms
from lbutils import BootstrapFormHelperMixin

from lbworkflow.forms import WorkflowFormMixin

from .models import Leave


class LeaveForm(BootstrapFormHelperMixin, WorkflowFormMixin, forms.ModelForm):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.init_crispy_helper()
        self.layout_fields([
            ['start_on', 'end_on'],
            ['leave_days', None],
            ['reason', ],
        ])

    def save(self, commit=True):
        obj = super().save(commit=False)
        obj.init_actual_info()
        if commit:
            self.save_m2m()
            obj.save()
        return obj

    class Meta:
        model = Leave
        fields = ['start_on', 'end_on', 'leave_days', 'reason', ]


class HRForm(BootstrapFormHelperMixin, WorkflowFormMixin, forms.ModelForm):
    comment = forms.CharField(
        label='Comment', required=False,
        widget=forms.Textarea())

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.init_crispy_helper(label_class='col-md-2', field_class='col-md-8')
        self.layout_fields([
            ['actual_start_on', ],
            ['actual_end_on', ],
            ['actual_leave_days', ],
            ['comment', ],
        ])

    class Meta:
        model = Leave
        fields = ['actual_start_on', 'actual_end_on', 'actual_leave_days', ]



from lbutils import as_callable

from lbworkflow import settings

# wf_send_sms(users, mail_type, event, ext_ctx)
# wf_send_mail(users, mail_type, event, ext_ctx)


def wf_send_msg(users, msg_type, event=None, ext_ctx=None):
    if not users:
        return

    users = set(users)
    if event:  # ignore operator
        if event.user in users:
            users = users.remove(event.user)

    for send_msg in settings.WF_SEND_MSG_FUNCS:
        as_callable(send_msg)(users, msg_type, event, ext_ctx)


def wf_print(users, msg_type, event=None, ext_ctx=None):
    print("wf_print: %s, %s, %s" % (users, msg_type, event))



import uuid

from django.conf import settings as django_settings
from django.db import models
from django.template import Context
from django.template import Template
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
from lbutils import create_instance
from lbutils import get_or_none

from lbworkflow import settings
from lbworkflow.core.helper import safe_eval

AUTH_USER_MODEL = getattr(django_settings, 'AUTH_USER_MODEL', 'auth.User')


class ProcessCategoryManager(models.Manager):
    def get_by_natural_key(self, uuid):
        return self.get(uuid=uuid)


class ProcessCategory(models.Model):
    uuid = models.UUIDField(unique=True, default=uuid.uuid4, editable=False)
    name = models.CharField(_('Name'), max_length=255, db_index=True)

    oid = models.IntegerField(_('Order'), default=999)
    is_active = models.BooleanField(_('Is active'), default=True)

    objects = ProcessCategoryManager()

    class Meta:
        ordering = ["oid"]

    def __str__(self):
        return self.name

    def natural_key(self):
        return (
            self.uuid,
        )

    def get_can_apply_processes(self, user, force_fetch=False):
        processes = getattr(self, '__cache__can_apply_processes', None)
        if processes and not force_fetch:
            return processes
        return self.process_set.all()

    def get_report_links(self):
        return ProcessReportLink.objects.filter(category=self)

    def get_all_process(self):
        return self.process_set.order_by('oid')


class ProcessReportLinkManager(models.Manager):
    def get_by_natural_key(self, uuid):
        return self.get(uuid=uuid)


class ProcessReportLink(models.Model):
    category = models.ForeignKey(
        ProcessCategory,
        blank=True, null=True,
        on_delete=models.SET_NULL, verbose_name='Category')
    name = models.CharField('Name', max_length=255)
    url = models.CharField('URL', max_length=255)
    open_in_new_window = models.BooleanField('Open in new window', default=False)
    perm = models.CharField(
        'Permission', max_length=255, blank=True,
        help_text="Permission to view this report")

    oid = models.IntegerField('Order', default=999)
    is_active = models.BooleanField('Is active', default=True)

    objects = ProcessReportLinkManager()

    class Meta:
        ordering = ["oid"]

    def __str__(self):
        return self.name

    def get_url(self):
        try:
            url_component_list = self.url.split(' ')
            return reverse(url_component_list[0], args=url_component_list[1:])
        except Exception:
            return self.url


class ProcessManager(models.Manager):
    def get_by_natural_key(self, code):
        return self.get(code=code)


class Process(models.Model):
    """
    A process holds the map that describes the flow of work.
    """
    code = models.CharField(
        'Code', max_length=100, unique=True,
        help_text='A unique code to identify process')
    prefix = models.CharField(
        'Prefix',
        max_length=8, default='', blank=True,
        help_text='prefix for process NO.')
    name = models.CharField(
        'Name', max_length=255,
        help_text='Name for this process')
    category = models.ForeignKey(
        ProcessCategory,
        blank=True, null=True,
        on_delete=models.SET_NULL,
        verbose_name='Category')
    note = models.TextField('Note', blank=True)

    oid = models.IntegerField('Order', default=999)
    is_active = models.BooleanField('Is active', default=True)
    ext_data = JSONField(default="{}")

    objects = ProcessManager()

    class Meta:
        verbose_name = 'Process'
        ordering = ["oid"]
        permissions = (
            ("sft_mgr_process", "workflow - Config"),
        )

    def natural_key(self):
        return (
            self.code,
        )

    def __str__(self):
        return self.name

    def have_apply_perm(self, user):
        # TODO check permission
        return True

    def get_draft_active(self):
        return get_or_none(Node, process=self, status='draft')

    def get_rejected_active(self):
        return get_or_none(Node, process=self, status='rejected')

    def get_given_up_active(self):
        return get_or_none(Node, process=self, status='given up')

    def get_rollback_transition(self, in_node, out_node):
        transition = Transition(
            name='Rollback',
            code='rollback',
            process=self,
            is_agree=False,
            can_auto_agree=False,
            input_node=in_node,
            output_node=out_node,
        )
        return transition

    def get_give_up_transition(self, in_node):
        output = self.get_given_up_active()
        transition = Transition(
            name='Give up',
            code='give up',
            process=self,
            is_agree=False,
            can_auto_agree=False,
            input_node=in_node,
            output_node=output,
        )
        return transition

    def get_back_to_transition(self, in_node, out_node=None):
        transition = Transition(
            name='Back to',
            code='back to',
            process=self,
            is_agree=False,
            can_auto_agree=False,
            input_node=in_node,
            output_node=out_node,
        )
        return transition

    def get_reject_transition(self, in_node):
        transition = Transition(
            name='Reject',
            code='reject',
            process=self,
            is_agree=False,
            can_auto_agree=False,
            input_node=in_node,
            output_node=self.get_rejected_active(),
        )
        return transition

    def get_add_assignee_transition(self, in_node):
        transition = Transition(
            name='Add assignee',
            code='add assignee',
            process=self,
            is_agree=False,
            can_auto_agree=False,
            input_node=in_node,
            output_node=in_node,
        )
        return transition


class NodeManager(models.Manager):
    def get_by_natural_key(self, uuid):
        return self.get(uuid=uuid)


class Node(models.Model):
    """
    Node is the states of an instance.
    """

    STATUS_CHOICES = (
        ('draft', 'Draft'),
        ('given up', 'Given up'),
        ('rejected', 'Rejected'),
        ('in progress', 'In Progress'),
        ('completed', 'Completed'),
    )
    AUDIT_PAGE_TYPE_CHOICES = (
        ('view', 'view'),
        ('edit', 'Edit'),
    )
    TYPE_CHOICES = (
        ('node', 'Node'),
        ('router', 'Router'),
    )

    uuid = models.UUIDField(unique=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey(
        'Process',
        on_delete=models.CASCADE,
        verbose_name='Process')
    name = models.CharField('Name', max_length=255)
    code = models.CharField(
        'Code', max_length=255, blank=True)
    step = models.IntegerField(
        'Step', default=0,
        help_text="",
    )
    status = models.CharField(
        'Status', max_length=16,
        default='in progress', choices=STATUS_CHOICES)
    audit_page_type = models.CharField(
        'Audit page type', max_length=64,
        choices=AUDIT_PAGE_TYPE_CHOICES,
        help_text='If this node can edit, will auto goto edit mode when audit.',
        default='view')
    node_type = models.CharField(
        'Status', max_length=16,
        default='node', choices=TYPE_CHOICES)

    can_edit = models.BooleanField('Can edit', default=False)
    can_reject = models.BooleanField('Can reject', default=True)
    can_give_up = models.BooleanField('Can give up', default=True)

    operators = models.TextField('Audit users', blank=True)
    notice_users = models.TextField('Notice users', blank=True)
    share_users = models.TextField('Share users', blank=True)

    note = models.TextField('Note', blank=True)
    # if not audit after xx day send a remind

    is_active = models.BooleanField('Is active', default=True)
    ext_data = JSONField(default="{}")

    objects = NodeManager()

    def __str__(self):
        return self.name

    def natural_key(self):
        return (
            self.uuid,
        )

    def is_submitted(self):
        return self.status in ['in progress', 'completed']

    def get_operators(self, owner, operator, instance=None):
        return create_instance(
            settings.USER_PARSER, self.operators, instance, operator, owner
        ).parse()

    def get_notice_users(self, owner, operator, instance=None):
        return create_instance(
            settings.USER_PARSER, self.notice_users, instance, operator, owner
        ).parse()

    def get_share_users(self, owner, operator, instance=None):
        return create_instance(
            settings.USER_PARSER, self.share_users, instance, operator, owner
        ).parse()

    def get_users(self, owner, operator, instance=None):
        operators = self.get_operators(owner, operator, instance)
        notice_users = self.get_notice_users(owner, operator, instance)
        share_users = self.get_share_users(owner, operator, instance)
        return operators, notice_users, share_users


class TransitionManager(models.Manager):
    def get_by_natural_key(self, uuid):
        return self.get(uuid=uuid)


class Transition(models.Model):
    """
    A Transition connects two node: a From and a To activity.
    """
    ROUTING_RULE_CHOICES = (
        ('split', 'split'),
        ('joint', 'Joint'),
    )
    uuid = models.UUIDField(unique=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey(
        'Process',
        on_delete=models.CASCADE,
        verbose_name='Process'
    )
    name = models.CharField(
        'Name', max_length=100, default='Agree',
        help_text="It also the action's name, like: Agree/Submit")
    code = models.CharField(  # 'back to', 'rollback'
        'Code', max_length=100, blank=True,
    )
    is_agree = models.BooleanField(
        'Is agree',
        default=True,
        help_text='User only need agree one time in one workflow'
    )
    can_auto_agree = models.BooleanField(
        'If can auto agree',
        default=True,
        help_text='If user agreed in previous steps will auto agree',
    )
    routing_rule = models.CharField(
        'Routing rule', max_length=16,
        default='split',
        choices=ROUTING_RULE_CHOICES,
        help_text="joint: do transition after all work item finished. joint: do transition immediately")
    input_node = models.ForeignKey(
        Node, verbose_name='Input node',
        null=True, on_delete=models.SET_NULL,
        related_name='input_transitions',
        help_text='')
    output_node = models.ForeignKey(
        Node, verbose_name='Output node',
        null=True, on_delete=models.SET_NULL,
        related_name='output_transitions',
        help_text='')
    app = models.ForeignKey(
        'App', verbose_name='Application to perform',
        null=True, on_delete=models.SET_NULL,
        blank=True)
    app_param = models.CharField(
        max_length=100, verbose_name='Param for application',
        null=True, blank=True,
        help_text='Depend on App config')
    condition = models.TextField(
        'Condition', blank=True,
        help_text="Uses the Python syntax.ex: `o.leave_days>3`")  # eval(t.condition, {'o': obj})
    note = models.TextField('Note', blank=True)

    oid = models.IntegerField(
        'Order', default=999)
    is_active = models.BooleanField('Is active', default=True)
    ext_data = JSONField(default="{}")

    objects = TransitionManager()

    def __str__(self):
        return '%s - %s' % (self.process.name, self.name)

    def natural_key(self):
        return (
            self.uuid,
        )

    def is_match_condition(self, wf_obj):
        if not self.condition.strip():
            return True
        return safe_eval(self.condition, {'o': wf_obj})

    def get_condition_descn(self):
        return self.condition.split('#')[-1].strip()

    def as_simple_agree_transition(self):
        self.pk = None
        if not self.code:
            self.code = 'agree'

    def get_app_url(self, task):
        return self.app.get_url(task, self)


class AppManager(models.Manager):
    def get_by_natural_key(self, uuid):
        return self.get(uuid=uuid)


class App(models.Model):
    """
    An application is a python view that can be called by URL.
    """
    APP_TYPE_CHOICES = (
        ('url', 'URL'),
    )
    uuid = models.UUIDField(unique=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100, help_text='Name')
    app_type = models.CharField('Type', max_length=255, choices=APP_TYPE_CHOICES, default='url')
    action = models.CharField(
        max_length=255, blank=True,
        help_text="URL: It can be url or django's url name. If it's blank will use transition's app param")
    note = models.TextField(blank=False)

    objects = AppManager()

    def __str__(self):
        return self.name

    def natural_key(self):
        return (
            self.uuid,
        )

    def get_url(self, task, transition):
        def render(templ_str, ctx):
            return Template(templ_str).render(Context(ctx))
        ts_id = transition.pk or transition.code

        ctx = {
            "wi": task,
            "wf_code": transition.process.code,
            "ts": transition,
            "ts_id": ts_id,
            "in": task.instance,
            "o": task.instance.content_object,
        }

        url = 'wf_process'
        if self.action:
            url = self.action
        elif transition.app_param:
            url = transition.app_param
        try:
            url_component_list = url.split(' ')
            url_param = [render(e, ctx) for e in url_component_list[1:]]
            url = reverse(url_component_list[0], args=url_param)
        except Exception:
            pass
        if "?" not in url:
            url = "%s?a=" % url
        url = "%s&ts_id={{ts_id}}&wi_id={{wi.id}}" % url
        return render(url, ctx)



from lbworkflow.views.generics import CreateView
from lbworkflow.views.generics import UpdateView
from lbworkflow.views.generics import WFListView

from .forms import SimpleWorkFlowForm
from .models import SimpleWorkFlow


class SimpleWorkFlowCreateView(CreateView):
    form_classes = {
        'form': SimpleWorkFlowForm,
    }


new = SimpleWorkFlowCreateView.as_view()


class SimpleWorkFlowUpdateView(UpdateView):
    form_classes = {
        'form': SimpleWorkFlowForm,
    }


edit = SimpleWorkFlowUpdateView.as_view()


class SimpleWorkFlowListView(WFListView):
    wf_code = 'simplewf'
    model = SimpleWorkFlow
    excel_file_name = 'simplewf'
    excel_titles = [
        'Created on', 'Created by',
        'Summary', 'Content',
        'Status',
    ]

    def get_excel_data(self, o):
        return [
            o.created_by.username, o.created_on,
            o.summary, o.content,
            o.pinstance.cur_node.name,
        ]


show_list = SimpleWorkFlowListView.as_view()



from setuptools import find_packages, setup
from lbworkflow import __version__

setup(
    name='django-lb-workflow',
    version=__version__,
    url='https://github.com/vicalloy/django-lb-workflow',
    author='vicalloy',
    author_email='vicalloy@gmail.com',
    description="Reusable workflow library for Django",
    license='BSD',
    packages=find_packages(exclude=['tests']),
    python_requires=">=3.5",
    include_package_data=True,
    install_requires=[
        'django>=2.2',
        'jsonfield>=1.0.1',
        'xlsxwriter>=0.9.6',
        'jinja2>=2.9.6',
        'django-lbutils>=1.1.0',
        'django-lbattachment>=1.1.0',
    ],
    tests_require=[
        'coverage',
        'flake8==3.7.9',
        'isort',
    ],
    extras_require={
        'options': [
            'django_select2>=7.2.0',
            'django-compressor>=2.1.1',
            'django-bower>=5.2.0',
            'django-crispy-forms>=1.6',
            'django-lb-adminlte>=1.1.0',
            'django-impersonate',
            'django-stronghold',
            'django-bootstrap-pagination>=1.7.0',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from django.db.models import Q
from lbutils import as_callable

from lbworkflow import settings


def safe_eval(source, globals, *args, **kwargs):
    globals['Q'] = Q
    for s in settings.EVAL_FUNCS:
        globals[s[0]] = as_callable(s[1])
    source = source.replace('import', '')
    return eval(source, globals, *args, **kwargs)



from django.contrib.auth import get_user_model
from django.urls import reverse

from lbworkflow.views.helper import user_wf_info_as_dict

from .leave.models import Leave
from .test_base import BaseTests

User = get_user_model()


class HelperTests(BaseTests):

    def test_user_wf_info_as_dict(self):
        leave = self.leave
        leave.submit_process()

        info = user_wf_info_as_dict(leave, self.users['tom'])
        self.assertIsNotNone(info['task'])
        self.assertIsNotNone(info['object'])
        self.assertFalse(info['can_give_up'])
        self.assertEqual(info['wf_code'], 'leave')

        info = user_wf_info_as_dict(leave, self.users['owner'])
        self.assertIsNone(info['task'])
        self.assertTrue(info['can_give_up'])

        info = user_wf_info_as_dict(leave, self.users['vicalloy'])
        self.assertIsNone(info['task'])


class ViewTests(BaseTests):

    def setUp(self):
        super().setUp()
        self.client.login(username='owner', password='password')

    def test_start_wf(self):
        resp = self.client.get(reverse('wf_start_wf'))
        self.assertEqual(resp.status_code, 200)

    def test_wf_list(self):
        resp = self.client.get(reverse('wf_list', args=('leave', )))
        self.assertEqual(resp.status_code, 200)

    def test_wf_report_list(self):
        resp = self.client.get(reverse('wf_report_list'))
        self.assertEqual(resp.status_code, 200)

    def test_wf_list_export(self):
        resp = self.client.get(reverse('wf_list', args=('leave', )), {'export': 1})
        self.assertEqual(resp.status_code, 200)

    def test_detail(self):
        resp = self.client.get(reverse('wf_detail', args=('1', )))
        self.assertEqual(resp.status_code, 200)
        self.assertContains(resp, 'from leave.detail')

    def test_submit(self):
        self.client.login(username='owner', password='password')

        url = reverse('wf_new', args=('leave', ))
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)

        data = {
            'start_on': '2017-04-19 09:01',
            'end_on': '2017-04-20 09:01',
            'leave_days': '1',
            'reason': 'test save',
        }
        resp = self.client.post(url, data)
        leave = Leave.objects.get(reason='test save')
        self.assertRedirects(resp, '/wf/%s/' % leave.pinstance.pk)
        self.assertEqual('Draft', leave.pinstance.cur_node.name)

        data['act_submit'] = 'Submit'
        data['reason'] = 'test submit'
        resp = self.client.post(url, data)
        leave = Leave.objects.get(reason='test submit')
        self.assertRedirects(resp, '/wf/%s/' % leave.pinstance.pk)
        self.assertEqual('A2', leave.pinstance.cur_node.name)

    def test_edit(self):
        self.client.login(username='owner', password='password')

        data = {
            'start_on': '2017-04-19 09:01',
            'end_on': '2017-04-20 09:01',
            'leave_days': '1',
            'reason': 'test save',
        }
        url = reverse('wf_new', args=('leave', ))
        resp = self.client.post(url, data)
        leave = Leave.objects.get(reason='test save')
        self.assertRedirects(resp, '/wf/%s/' % leave.pinstance.pk)
        self.assertEqual('Draft', leave.pinstance.cur_node.name)

        url = reverse('wf_edit', args=(leave.pinstance.pk, ))
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)

        data['act_submit'] = 'Submit'
        data['reason'] = 'test submit'
        resp = self.client.post(url, data)
        leave = Leave.objects.get(reason='test submit')
        self.assertRedirects(resp, '/wf/%s/' % leave.pinstance.pk)
        self.assertEqual('A2', leave.pinstance.cur_node.name)

    def test_delete(self):
        self.client.login(username='admin', password='password')
        # POST
        url = reverse('wf_delete')
        leave = self.create_leave('to delete')
        data = {'pk': leave.pinstance.pk}
        resp = self.client.post(url, data)
        self.assertRedirects(resp, '/wf/list/')
        self.assertIsNone(self.get_leave('to delete'))

        # GET
        leave = self.create_leave('to delete')
        data = {'pk': leave.pinstance.pk}
        resp = self.client.get(url, data)
        self.assertRedirects(resp, '/wf/list/')
        self.assertIsNone(self.get_leave('to delete'))



from lbworkflow.core.datahelper import create_category
from lbworkflow.core.datahelper import create_node
from lbworkflow.core.datahelper import create_process
from lbworkflow.core.datahelper import create_transition


def load_data():
    load_issue()


def load_issue():
    """ load_[wf_code] """
    category = create_category('5f31d065-00cc-0020-beea-641f0a670010', 'HR')
    process = create_process('purchase', 'Purchase', category=category)
    create_node('5f31d065-00a0-0030-beea-641f0a670010', process, 'Draft', status='draft')
    create_node('5f31d065-00a0-0030-beea-641f0a670020', process, 'Given up', status='given up')
    create_node('5f31d065-00a0-0030-beea-641f0a670030', process, 'Rejected', status='rejected')
    create_node('5f31d065-00a0-0030-beea-641f0a670040', process, 'Completed', status='completed')
    create_node('5f31d065-00a0-0030-beea-641f0a670050', process, 'A1', operators='[owner]')
    create_transition('5f31d065-00e0-0030-beea-641f0a670010', process, 'Draft,', 'A1')
    create_transition('5f31d065-00e0-0030-beea-641f0a670020', process, 'A1,', 'Completed')



import inspect
import os
import shutil
import stat

from jinja2 import Environment
from jinja2 import FileSystemLoader

__all__ = (
    'FlowAppGenerator', 'clean_generated_files'
)


def clean_generated_files(model_class):
    folder_path = os.path.dirname(inspect.getfile(model_class))
    for path, dirs, files in os.walk(folder_path):
        if not path.endswith(model_class.__name__.lower()):
            shutil.rmtree(path)
        for file in files:
            if file not in ['models.py', 'wfdata.py', '__init__.py']:
                try:
                    os.remove(os.path.join(path, file))
                except:  # NOQA
                    pass


def get_fields(model_class):
    fields = []
    ignore_fields = ['id', 'pinstance', 'created_on', 'created_by']
    for f in model_class._meta.fields:
        if f.name not in ignore_fields:
            fields.append(f)
    return fields


def get_field_names(model_class):
    fields = get_fields(model_class)
    return ', '.join(["'%s'" % e.name for e in fields])


def group(flat_list):
    for i in range(len(flat_list) % 2):
        flat_list.append(None)
        pass
    return list(zip(flat_list[0::2], flat_list[1::2]))


class FlowAppGenerator(object):
    def __init__(self, app_template_path=None):
        if not app_template_path:
            app_template_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'app_template')
        self.app_template_path = app_template_path
        super().__init__()

    def init_env(self, template_path):
        loader = FileSystemLoader(template_path)
        self.env = Environment(
            block_start_string='[%',
            block_end_string='%]',
            variable_start_string='[[',
            variable_end_string=']]',
            comment_start_string='[#',
            comment_end_string='#]',
            loader=loader,
        )

    def gen(self, model_class, item_model_class_list=None, wf_code=None, replace=False, ignores=['wfdata.py']):
        dest = os.path.dirname(inspect.getfile(model_class))
        app_name = model_class.__module__.split('.')[-2]
        if not wf_code:
            wf_code = app_name
        ctx = {
            'app_name': app_name,
            'wf_code': wf_code,
            'class_name': model_class.__name__,
            'wf_name': model_class._meta.verbose_name,
            'field_names': get_field_names(model_class),
            'fields': get_fields(model_class),
            'grouped_fields': group(get_fields(model_class)),
        }
        if item_model_class_list:
            item_list = []
            for item_model_class in item_model_class_list:
                item_ctx = {
                    'class_name': item_model_class.__name__,
                    'lowercase_class_name': item_model_class.__name__.lower(),
                    'field_names': get_field_names(item_model_class),
                    'fields': get_fields(item_model_class),
                    'grouped__fields': group(get_fields(item_model_class)),
                }
                item_list.append(item_ctx)
            ctx['item_list'] = item_list
        self.copy_template(self.app_template_path, dest, ctx, replace, ignores)

    def copy_template(self, src, dest, ctx={}, replace=False, ignores=[]):
        self.init_env(src)
        for path, dirs, files in os.walk(src):
            relative_path = path[len(src):].lstrip(os.path.sep)
            dest_path = os.path.join(dest, relative_path)
            dest_path = dest_path.replace('app_name', ctx.get('app_name', 'app_name'))
            if not os.path.exists(dest_path):
                os.mkdir(dest_path)
            for i, subdir in enumerate(dirs):
                if subdir.startswith('.'):
                    del dirs[i]
            for filename in files:
                if filename.endswith('.pyc') or filename.startswith('.'):
                    continue
                src_file_path = os.path.join(path, filename)
                src_file_path = src_file_path[len(src):].strip(os.path.sep)
                dest_file_path = os.path.join(dest, relative_path, filename)
                dest_file_path = dest_file_path.replace('app_name', ctx.get('app_name', 'app_name'))
                if dest_file_path.endswith('-tpl'):
                    dest_file_path = dest_file_path[:-4]

                is_exists = os.path.isfile(dest_file_path)
                for ignore in ignores:
                    if dest_file_path.endswith(ignore):
                        replace = False
                if is_exists and not replace:
                    continue
                self.copy_template_file(src_file_path, dest_file_path, ctx)

    def copy_template_file(self, src, dest, ctx={}):
        if os.path.sep != '/':
            # https://github.com/pallets/jinja/issues/767
            # Jinja template names are not fileystem paths.
            # They always use forward slashes so this is working as intended.
            src = src.replace(os.path.sep, '/')
        template = self.env.get_template(src)
        template.stream(ctx).dump(dest, encoding='utf-8')
        # Make new file writable.
        if os.access(dest, os.W_OK):
            st = os.stat(dest)
            new_permissions = stat.S_IMODE(st.st_mode) | stat.S_IWUSR
            os.chmod(dest, new_permissions)
